"""Methods to handle MATLAB FileWrapper__ data"""

import warnings

import numpy as np

from matio.utils.matclass import (
    DYNAMIC_PROPERTY_PREFIX,
    ENUMERATION_INSTANCE_TAG,
    MCOS_SUBSYSTEM_CLASS,
    EmptyMatStruct,
    MatlabCanonicalEmpty,
    MatlabEnumerationArray,
    MatlabOpaque,
    MatlabOpaqueArray,
    MatlabOpaqueProperty,
    MatReadError,
    MatReadWarning,
    MatWriteWarning,
    OpaqueType,
    PropertyType,
)
from matio.utils.matconvert import (
    ENUM_INSTANCE_DTYPE,
    convert_mat_to_py,
    enum_to_opaque,
    mat_to_enum,
    matlab_classdef_types,
    matlab_saveobj_ret_types,
)
from matio.utils.matheaders import MAT_HDF_VERSION, MCOS_MAGIC_NUMBER
from matio.utils.matutils import to_writeable

SYSTEM_BYTE_ORDER = "<" if np.little_endian else ">"

FILEWRAPPER_VERSION = 4
MIN_FILEWRAPPER_VERSION = 2


class MatSubsystem:
    """Representation class for MATLAB FileWrapper__ data"""

    def __init__(
        self,
        byte_order,
        raw_data=False,
        add_table_attrs=False,
        oned_as="col",
        saveobj_classes=[],
    ):
        self.byte_order = "<u4" if byte_order[0] == "<" else ">u4"
        self.raw_data = raw_data
        self.add_table_attrs = add_table_attrs
        self.oned_as = oned_as

        self.mcos_object_cache = {}

        self.version = None
        self.num_names = 0
        self.mcos_names = []

        # Metadata Regions
        self.class_id_metadata = []
        self.object_id_metadata = []
        self.saveobj_metadata = []
        self.nobj_metadata = []
        self.dynprop_metadata = []
        self._u6_metadata = None  # Unknown Object Metadata
        self._u7_metadata = None  # Unknown Object Metadata

        self.mcos_props_saved = []

        # Class Template Data
        self._c3 = None  # Unknown Class Template
        self.mcos_class_alias_metadata = []
        self.mcos_props_defaults = None

        self._handle_data = None
        self._java_data = None

        # Counters for object serialization
        self.saveobj_counter = 0
        self.nobj_counter = 0
        self.class_id_counter = 0
        self.object_id_counter = 0

        self.saveobj_class_names = matlab_saveobj_ret_types + saveobj_classes

    def init_save(self):
        """Initializes save with metadata for object ID = 0"""

        self.version = FILEWRAPPER_VERSION
        self.class_id_metadata.extend([0, 0, 0, 0])
        self.dynprop_metadata.extend([0, 0])
        self.mcos_class_alias_metadata.append(0)

        # These metadata fields are written per object before unrolling
        # nested properties. Each list represents one object which is mutable
        # This does not matter for class ID
        self.object_id_metadata.append([0, 0, 0, 0, 0, 0])
        self.saveobj_metadata.append([0, 0])
        self.nobj_metadata.append([0, 0])

    def load_subsystem(self, subsystem_data):
        """Parse and cache subsystem data"""

        for field in subsystem_data.dtype.names:
            if field == OpaqueType.JAVA:
                self._java_data = subsystem_data[0, 0][field]
            if field == OpaqueType.HANDLE:
                self._handle_data = subsystem_data[0, 0][field]
            if field == OpaqueType.MCOS:
                self.load_mcos_data(subsystem_data[0, 0][field])

    def load_fwrap_metadata(self, fwrap_metadata):
        """Parse and cache FileWrapper__ metadata"""

        self.version = np.frombuffer(
            fwrap_metadata, dtype=self.byte_order, count=1, offset=0
        )[0]
        if not MIN_FILEWRAPPER_VERSION <= self.version <= FILEWRAPPER_VERSION:
            raise MatReadError(
                f"{MCOS_SUBSYSTEM_CLASS} version {self.version} is not supported"
            )

        # Number of unique property and class names
        self.num_names = np.frombuffer(
            fwrap_metadata, dtype=self.byte_order, count=1, offset=4
        )[0]

        # 8 offsets to different regions within this cell
        region_offsets = np.frombuffer(
            fwrap_metadata, dtype=self.byte_order, count=8, offset=8
        )

        # A list of null terminated Property and Class Names
        data = fwrap_metadata[40 : region_offsets[0]].tobytes()
        raw_strings = data.split(b"\x00")
        self.mcos_names = [s.decode("ascii") for s in raw_strings if s]

        # Region 1: Class ID Metadata
        self.class_id_metadata = np.frombuffer(
            fwrap_metadata,
            dtype=self.byte_order,
            count=(region_offsets[1] - region_offsets[0]) // 4,
            offset=region_offsets[0],
        )

        # Region 2: Saveobj Prop Metadata
        self.saveobj_metadata = np.frombuffer(
            fwrap_metadata,
            dtype=self.byte_order,
            count=(region_offsets[2] - region_offsets[1]) // 4,
            offset=region_offsets[1],
        )

        # Region 3: Object ID Metadata
        self.object_id_metadata = np.frombuffer(
            fwrap_metadata,
            dtype=self.byte_order,
            count=(region_offsets[3] - region_offsets[2]) // 4,
            offset=region_offsets[2],
        )

        # Region 4: Object Prop Metadata
        self.nobj_metadata = np.frombuffer(
            fwrap_metadata,
            dtype=self.byte_order,
            count=(region_offsets[4] - region_offsets[3]) // 4,
            offset=region_offsets[3],
        )

        # Region 5: Dynamic Prop Metadata
        self.dynprop_metadata = np.frombuffer(
            fwrap_metadata,
            dtype=self.byte_order,
            count=(region_offsets[5] - region_offsets[4]) // 4,
            offset=region_offsets[4],
        )

        # Following may be reserved in some versions
        # Unknown data, kept raw
        if region_offsets[6] > 0:
            self._u6_metadata = fwrap_metadata[region_offsets[5] : region_offsets[6]]
        if region_offsets[7] > 0:
            self._u7_metadata = fwrap_metadata[region_offsets[6] : region_offsets[7]]

    def load_mcos_data(self, fwrap_data):
        """Parse and cache MCOS FileWrapper__ data"""

        fwrap_metadata = fwrap_data[0, 0]
        self.load_fwrap_metadata(fwrap_metadata)

        if self.version == 2:
            self.mcos_props_saved = fwrap_data[2:-1, 0]
        elif self.version == 3:
            self.mcos_props_saved = fwrap_data[2:-2, 0]
            self.mcos_class_alias_metadata = fwrap_data[-2, 0]
        else:
            self.mcos_props_saved = fwrap_data[2:-3, 0]
            self._c3 = fwrap_data[-3, 0]
            self.mcos_class_alias_metadata = fwrap_data[-2, 0]

        self.mcos_props_defaults = fwrap_data[-1, 0]

    def is_valid_mcos_enumeration(self, metadata):
        """Checks if property value is a valid MCOS enumeration metadata array"""

        if metadata.dtype.names:
            if ENUMERATION_INSTANCE_TAG in metadata.dtype.names:
                if (
                    metadata[0, 0][ENUMERATION_INSTANCE_TAG].dtype == np.uint32
                    and metadata[0, 0][ENUMERATION_INSTANCE_TAG].size == 1
                    and metadata[0, 0][ENUMERATION_INSTANCE_TAG] == MCOS_MAGIC_NUMBER
                ):
                    return True

        return False

    def is_valid_mcos_object(self, metadata):
        """Checks if property value is a valid MCOS metadata array"""

        if not (
            metadata.dtype == np.uint32
            and metadata.ndim == 2
            and metadata.shape[1] == 1
            and metadata.size >= 3
        ):
            return False

        if metadata[0, 0] != MCOS_MAGIC_NUMBER:
            return False

        return True

    def is_valid_opaque_object(self, metadata):
        """Checks if property value is a valid opaque object metadata array"""

        # Only know MCOS identifier
        # Can include other types later
        return self.is_valid_mcos_object(metadata)

    def check_prop_for_opaque(self, prop):
        """Check if a property value in FileWrapper__ contains opaque objects during load"""

        if not isinstance(prop, np.ndarray):
            return prop

        if prop.dtype.hasobject:
            if prop.dtype.names:
                # Iterate though struct array
                # Also handles MatlabObject, MatlabFunction
                if self.is_valid_mcos_enumeration(prop):
                    return self.load_mcos_enumeration(prop, type_system=OpaqueType.MCOS)
                else:
                    for idx in np.ndindex(prop.shape):
                        for name in prop.dtype.names:
                            field_val = prop[idx][name]
                            prop[idx][name] = self.check_prop_for_opaque(field_val)

            else:
                # Iterate through cell arrays
                # NOTE: Function Handles to classdef methods have an MCOS identifier
                # But I don't think there's anything to read them as opaque objects
                for idx in np.ndindex(prop.shape):
                    cell_item = prop[idx]
                    prop[idx] = self.check_prop_for_opaque(cell_item)

        elif self.is_valid_opaque_object(prop):
            # MCOS class names are derived from subsystem
            # For other types, it may be derived from metadata instead
            # So we use a placeholder classname
            prop = self.load_opaque_object(prop, type_system=OpaqueType.MCOS)

        return prop

    def get_class_alias(self, class_id):
        """Extracts class alias for a given object from its class ID."""

        class_alias_idx = self.mcos_class_alias_metadata[class_id].item()
        if class_alias_idx != 0:
            class_alias = self.mcos_names[class_alias_idx - 1]
        else:
            class_alias = None

        return class_alias

    def get_classname(self, class_id):
        """Extracts class name with namespace qualifier for a given object from its class ID."""

        namespace_idx = self.class_id_metadata[class_id * 4]
        classname_idx = self.class_id_metadata[class_id * 4 + 1]
        # Remaining two fields are unknowns

        if namespace_idx == 0:
            namespace = ""
        else:
            namespace = self.mcos_names[namespace_idx - 1] + "."

        classname = namespace + self.mcos_names[classname_idx - 1]
        return classname

    def get_object_metadata(self, object_id):
        """Extracts object dependency IDs for a given object."""

        class_id, _, _, saveobj_id, normobj_id, dep_id = self.object_id_metadata[
            object_id * 6 : object_id * 6 + 6
        ]
        # Ignored fields are unknowns
        return class_id, saveobj_id, normobj_id, dep_id

    def get_default_properties(self, class_id):
        """Returns the default properties (as dict) for a given class ID"""

        prop_arr = self.mcos_props_defaults[class_id, 0]
        prop_map = {}
        if prop_arr.dtype.names:
            for prop_name in prop_arr.dtype.names:
                prop_map[prop_name] = self.check_prop_for_opaque(
                    prop_arr[prop_name][0, 0]
                )

        return prop_map

    def get_property_idxs(self, obj_type_id, saveobj_ret_type):
        """Returns the property (name, type, value) metadata for an object ID"""

        prop_field_idxs = (
            self.saveobj_metadata if saveobj_ret_type else self.nobj_metadata
        )

        nfields = 3
        offset = prop_field_idxs[0]
        for _ in range(obj_type_id):
            nprops = prop_field_idxs[offset]
            offset += 1 + nfields * nprops
            offset += offset % 2  # Padding

        nprops = prop_field_idxs[offset]
        offset += 1
        return prop_field_idxs[offset : offset + nprops * nfields].reshape(
            (nprops, nfields)
        )

    def get_saved_properties(self, obj_type_id, saveobj_ret_type):
        """Returns the saved properties (as dict) for an object ID"""

        save_prop_map = {}

        prop_field_idxs = self.get_property_idxs(obj_type_id, saveobj_ret_type)
        for prop_idx, prop_type, prop_value in prop_field_idxs:
            prop_name = self.mcos_names[prop_idx - 1]
            if prop_type == PropertyType.MATLAB_ENUMERATION:
                save_prop_map[prop_name] = MatlabOpaqueProperty(
                    self.mcos_names[prop_value - 1],
                    ptype=PropertyType.MATLAB_ENUMERATION,
                )
            elif prop_type == PropertyType.PROPERTY_VALUE:
                save_prop_map[prop_name] = self.check_prop_for_opaque(
                    self.mcos_props_saved[prop_value]
                )
            elif prop_type == PropertyType.INTEGER_VALUE:
                save_prop_map[prop_name] = MatlabOpaqueProperty(
                    prop_value, ptype=PropertyType.INTEGER_VALUE
                )
            else:
                warnings.warn(
                    f'Unknown property type {prop_type} for property "{prop_name}"',
                    MatReadWarning,
                    stacklevel=3,
                )
                save_prop_map[prop_name] = MatlabOpaqueProperty(
                    prop_value, ptype=prop_type
                )

        return save_prop_map

    def get_dynamic_properties(self, dep_id):
        """Returns dynamicproperties (as dict) for a given object based on dependency ID"""

        offset = self.dynprop_metadata[0]
        for i in range(dep_id):
            nprops = self.dynprop_metadata[offset]
            offset += 1 + nprops
            offset += offset % 2  # Padding

        ndynprops = self.dynprop_metadata[offset]
        offset += 1
        dyn_prop_type2_ids = self.dynprop_metadata[offset : offset + ndynprops]

        if ndynprops == 0:
            return {}

        dyn_prop_map = {}
        for i, dyn_prop_id in enumerate(dyn_prop_type2_ids):
            dyn_class_id = self.get_object_metadata(dyn_prop_id)[0]
            classname = self.get_classname(dyn_class_id)
            dynobj = MatlabOpaque(
                properties=None,
                classname=classname,
                type_system=OpaqueType.MCOS,
                class_alias=self.get_class_alias(dyn_class_id),
            )
            dynobj.properties = self.get_properties(dyn_prop_id)
            dyn_prop_map[f"{DYNAMIC_PROPERTY_PREFIX}{i + 1}"] = dynobj

        return dyn_prop_map

    def get_properties(self, object_id):
        """Returns the properties as a dict for a given object ID"""

        if object_id == 0:
            # Matlab uses an object ID=0
            # MATLAB seems to keep references to deleted objects
            # Observed this in fig files I think? Don't remember
            # objectID=0 may be a placeholder for such cases
            return None

        class_id, saveobj_id, normobj_id, dep_id = self.get_object_metadata(object_id)
        if saveobj_id != 0:
            saveobj_ret_type = True
            obj_type_id = saveobj_id
        else:
            saveobj_ret_type = False
            obj_type_id = normobj_id

        prop_map = self.get_default_properties(class_id)
        prop_map.update(self.get_saved_properties(obj_type_id, saveobj_ret_type))
        prop_map.update(self.get_dynamic_properties(dep_id))

        return prop_map

    def set_mcos_name(self, name):
        """Gets or creates index for a name in mcos_names"""
        try:
            return self.mcos_names.index(name) + 1  # 1-based
        except ValueError:
            # Name doesn't exist, add it
            self.mcos_names.append(name)
            self.num_names += 1
            return self.num_names

    def set_class_id(self, classname, class_alias=None):
        """Sets the class ID for a given class name (including namespace)"""

        for class_id in range(1, self.class_id_counter + 1):
            existing_name = self.get_classname(class_id)
            if existing_name == classname:
                return class_id

        self.class_id_counter += 1

        # Add new class ID metadata
        namespace, _, cname = classname.rpartition(".")
        cname_idx = self.set_mcos_name(cname)
        if namespace:
            namespace_idx = self.set_mcos_name(namespace)
        else:
            namespace_idx = 0

        metadata = [namespace_idx, cname_idx, 0, 0]
        self.class_id_metadata.extend(metadata)

        class_alias_idx = self.set_mcos_name(class_alias) if class_alias else 0
        self.mcos_class_alias_metadata.append(class_alias_idx)

        return self.class_id_counter

    def serialize_nested_props(self, prop_value):
        """Recursively serializes nested properties of an MCOS object"""

        if isinstance(prop_value, MatlabEnumerationArray):
            prop_value = self.set_enumeration_metadata(prop_value)

        elif (
            isinstance(prop_value, np.ndarray)
            and prop_value.dtype.hasobject
            and prop_value.dtype.kind != "T"
        ):
            # Iterate through cell arrays and struct arrays
            # Ignore StringDType(): Used for MATLAB strings which uses 'O' internally

            # Make a copy to avoid modifying the original
            # Should be cheap as we're only copying object arrays
            prop_value = prop_value.copy()

            if prop_value.dtype.names:
                # Struct array
                for idx in np.ndindex(prop_value.shape):
                    for name in prop_value.dtype.names:
                        field_val = prop_value[idx][name]
                        field_val = to_writeable(field_val)
                        prop_value[idx][name] = self.serialize_nested_props(field_val)
            else:
                # Cell array
                for idx in np.ndindex(prop_value.shape):
                    cell_item = prop_value[idx]
                    cell_item = to_writeable(cell_item)
                    prop_value[idx] = self.serialize_nested_props(cell_item)

        else:
            prop_value = to_writeable(prop_value)

        if isinstance(prop_value, (MatlabOpaque, MatlabOpaqueArray)):
            prop_value = self.set_object_metadata(prop_value)

        return prop_value

    def serialize_object_props(self, prop_map, obj_prop_metadata):
        """Serializes the properties of an MCOS object"""

        object_id = self.object_id_counter
        nprops = len(prop_map)
        obj_prop_metadata.extend([nprops])

        dynprop_ids = [0]

        for prop_name, prop_value in prop_map.items():
            if prop_name.startswith(DYNAMIC_PROPERTY_PREFIX):
                dynobj_id, _ = self.set_object_id(prop_value)
                dynprop_ids.append(dynobj_id)
                obj_prop_metadata[0] -= 1
                # Dynamic Props are not counted as class props
                continue

            if prop_name[0] in "_0123456789":
                msg = f"Property names cannot start with '{prop_name[0]}'. Skipping '{prop_name}'"
                warnings.warn(msg, MatWriteWarning, stacklevel=3)
                continue

            field_name_idx = self.set_mcos_name(prop_name)
            prop_value = self.serialize_nested_props(prop_value)

            prop_vals = [field_name_idx, 0, 0]
            if isinstance(prop_value, MatlabOpaqueProperty):
                if prop_value.ptype == PropertyType.MATLAB_ENUMERATION:
                    prop_vals[1] = PropertyType.MATLAB_ENUMERATION
                    prop_vals[2] = self.set_mcos_name(prop_value.item())
                elif prop_value.ptype == PropertyType.INTEGER_VALUE:
                    prop_vals[1] = PropertyType.INTEGER_VALUE
                    prop_vals[2] = prop_value.item()
                else:
                    warnings.warn(
                        f"Unknown property type {prop_value.ptype} for property '{prop_name}'. Saving as raw integer value",
                        MatWriteWarning,
                        stacklevel=3,
                    )
                    prop_vals[1] = prop_value.ptype
                    prop_vals[2] = prop_value.item()
            else:
                cell_number_idx = len(self.mcos_props_saved)
                self.mcos_props_saved.append(prop_value)
                prop_vals[1] = PropertyType.PROPERTY_VALUE
                prop_vals[2] = cell_number_idx

            obj_prop_metadata.extend(prop_vals)

        if len(obj_prop_metadata) % 2 == 1:
            obj_prop_metadata.append(0)  # Padding

        if len(dynprop_ids) > 1:
            ndynprops = len(dynprop_ids) - 1
            dynprop_ids[0] = ndynprops
            if len(dynprop_ids) % 2 == 1:
                dynprop_ids.append(0)  # Padding
        else:
            dynprop_ids.append(0)  # No dynamic props
        self.dynprop_metadata.extend(dynprop_ids)

        ndeps = self.object_id_counter - object_id
        return ndeps

    def set_object_id(self, obj, saveobj_ret_type=False):
        """Sets the object ID for a given object key (id(obj) or obj_key)"""

        if obj.properties is None:
            # Deleted object
            class_id = self.set_class_id(obj.classname, obj.class_alias)
            obj_id = 0
            return obj_id, class_id

        if obj in self.mcos_object_cache:
            class_id = self.set_class_id(obj.classname, obj.class_alias)
            return self.mcos_object_cache[obj], class_id

        self.object_id_counter += 1
        mat_obj_id = self.object_id_counter
        self.mcos_object_cache[obj] = mat_obj_id

        obj_prop_metadata = []
        saveobj_id = 0
        nobj_id = 0
        if saveobj_ret_type:
            self.saveobj_counter += 1
            saveobj_id = self.saveobj_counter
            self.saveobj_metadata.append(obj_prop_metadata)
        else:
            self.nobj_counter += 1
            nobj_id = self.nobj_counter
            self.nobj_metadata.append(obj_prop_metadata)

        obj_id_metadata = [0] * 6
        self.object_id_metadata.append(obj_id_metadata)

        # * Can we optimize here?
        if saveobj_ret_type and not (
            len(obj.properties) == 1 and set(obj.properties.keys()) == {"any"}
        ):
            raise MatReadError(
                f"Object of class {obj.classname} marked with a saveobj return type must have a single property 'any' containing the return value of its saveobj method"
            )

        ndeps = self.serialize_object_props(obj.properties, obj_prop_metadata)

        class_id = self.set_class_id(obj.classname, obj.class_alias)
        obj_id_metadata[0] = class_id
        if saveobj_ret_type:
            obj_id_metadata[3] = saveobj_id
        else:
            obj_id_metadata[4] = nobj_id

        obj_id_metadata[5] = mat_obj_id + ndeps
        for i in range(1, ndeps + 1):
            obj_id = self.object_id_counter - ndeps + i
            self.object_id_metadata[obj_id][5] -= 1

        return mat_obj_id, class_id

    def set_mcos_object_metadata(self, obj):
        """Sets metadata for a single MCOS object"""

        arr_ids = []
        classname = obj.classname
        saveobj_ret_type = classname in self.saveobj_class_names

        if isinstance(obj, MatlabOpaqueArray):
            for obj_elem in obj.ravel(order="F"):
                object_id, class_id = self.set_object_id(obj_elem, saveobj_ret_type)
                arr_ids.append(object_id)
            dims = obj.shape
        else:
            if isinstance(obj.properties, tuple):
                # 0x0, 1x0, 0x1 objects
                dims = obj.properties
                class_id = self.set_class_id(classname, obj.class_alias)
            else:
                object_id, class_id = self.set_object_id(obj, saveobj_ret_type)
                arr_ids.append(object_id)
                dims = (1, 1)

        return self.create_mcos_metadata(dims, arr_ids, class_id)

    def create_mcos_metadata(self, dims, arr_ids, class_id):
        """Creates MCOS metadata array"""

        ndims = len(dims)
        values = [MCOS_MAGIC_NUMBER, ndims] + list(dims) + list(arr_ids) + [class_id]
        return np.array(values, dtype=np.uint32).reshape(-1, 1)

    def set_object_metadata(self, obj):
        """Sets metadata for a MatioOpaque object"""

        type_system = obj.type_system
        if type_system != OpaqueType.MCOS:
            warnings.warn(
                "subsystem:set_object_metadata: Only MCOS objects are supported currently. This item will be skipped",
                MatWriteWarning,
            )
            return np.empty((0, 0), dtype=np.uint8)

        return self.set_mcos_object_metadata(obj)

    def set_enumeration_metadata(self, enum_array):
        """Creates MCOS Enumeration metadata array"""

        classname = enum_array.classname
        classname_idx = self.set_class_id(classname)

        value_names_idx = []
        values = []
        value_indices = []

        for i, enum_member in enumerate(enum_array.ravel(order="F")):
            value_name = enum_member.name

            value_class = enum_to_opaque(classname, enum_member.value)
            values.append(self.set_object_metadata(value_class))
            value_indices.append(i)
            value_names_idx.append(self.set_mcos_name(value_name))

        prop_name = next(iter(enum_array[0, 0].value))
        builtin_classname, _, _ = prop_name.rpartition(".")
        if builtin_classname:
            builtin_classname_idx = self.set_class_id(builtin_classname)
        else:
            builtin_classname_idx = 0

        value_names_idx = np.array(value_names_idx, dtype=np.uint32).reshape(-1, 1)
        values_arr = np.empty((len(values), 1), dtype=object)
        values_arr[:, 0] = values
        value_indices = np.array(value_indices, dtype=np.uint32).reshape(
            enum_array.shape, order="F"
        )

        enum_instance_metadata = np.empty((1, 1), dtype=ENUM_INSTANCE_DTYPE)
        enum_instance_metadata[0, 0]["EnumerationInstanceTag"] = np.array(
            MCOS_MAGIC_NUMBER, dtype=np.uint32
        ).reshape(1, 1)
        enum_instance_metadata[0, 0]["ClassName"] = np.array(
            classname_idx, dtype=np.uint32
        ).reshape(1, 1)
        enum_instance_metadata[0, 0]["BuiltinClassName"] = np.array(
            builtin_classname_idx, dtype=np.uint32
        ).reshape(1, 1)
        enum_instance_metadata[0, 0]["ValueNames"] = value_names_idx
        enum_instance_metadata[0, 0]["Values"] = values_arr
        enum_instance_metadata[0, 0]["ValueIndices"] = value_indices

        return enum_instance_metadata

    def set_fwrap_metadata(self):
        """Create FileWrapper Metadata Array"""

        regions = []
        regions.append(np.array([self.version], dtype=np.uint32).view(np.uint8))
        regions.append(np.array([self.num_names], dtype=np.uint32).view(np.uint8))

        # Region offsets (uint32 array)
        region_offsets = np.zeros(8, dtype=np.uint32)
        regions.append(region_offsets.view(np.uint8))

        # Names string (null-terminated)
        names_bytes = (
            b"\x00".join(name.encode("ascii") for name in self.mcos_names) + b"\x00"
        )
        pad_len = (8 - len(names_bytes) % 8) % 8
        names_bytes += b"\x00" * pad_len
        names_uint8 = np.frombuffer(names_bytes, dtype=np.uint8)
        regions.append(names_uint8)
        region_offsets[0] = 40 + names_uint8.size

        # Region 1 - Class ID Metadata
        region1 = np.array(self.class_id_metadata, dtype=np.uint32).view(np.uint8)
        regions.append(region1)
        region_offsets[1] = region_offsets[0] + region1.size

        # Region 2 - Saveobj Metadata
        self.saveobj_metadata = [x for sub in self.saveobj_metadata for x in sub]
        region2 = (
            np.array(self.saveobj_metadata, dtype=np.uint32).view(np.uint8)
            if len(self.saveobj_metadata) > 2
            else np.empty(0, dtype=np.uint8)
        )
        regions.append(region2)
        region_offsets[2] = region_offsets[1] + region2.size

        # Region 3 - Object ID Metadata
        self.object_id_metadata = [x for sub in self.object_id_metadata for x in sub]
        region3 = np.array(self.object_id_metadata, dtype=np.uint32).view(np.uint8)
        regions.append(region3)
        region_offsets[3] = region_offsets[2] + region3.size

        # Region 4 - Object Prop Metadata
        self.nobj_metadata = [x for sub in self.nobj_metadata for x in sub]
        region4 = (
            np.array(self.nobj_metadata, dtype=np.uint32).view(np.uint8)
            if len(self.nobj_metadata) > 0
            else np.empty(0, dtype=np.uint8)
        )
        regions.append(region4)
        region_offsets[4] = region_offsets[3] + region4.size

        # Region 5 - Dynamic Prop Metadata
        region5 = np.array(self.dynprop_metadata, dtype=np.uint32).view(np.uint8)
        regions.append(region5)
        region_offsets[5] = region_offsets[4] + region5.size

        # Region 6 - Unknown
        region6 = np.empty(0, dtype=np.uint8)
        regions.append(region6)
        region_offsets[6] = region_offsets[5] + region6.size

        # Region 7 - Unknown
        region7 = np.zeros(1, np.uint64).view(np.uint8)
        regions.append(region7)
        region_offsets[7] = region_offsets[6] + region7.size

        fwrap_metadata = np.concatenate(regions)
        return fwrap_metadata.reshape(-1, 1)

    def set_fwrap_data(self):
        """Create FileWrapper Cell Array"""

        if self.class_id_counter == 0:
            # No MCOS objects to save
            return None

        array_len = 5 + len(self.mcos_props_saved)
        fwrap_data = np.empty((array_len, 1), dtype=object)

        fwrap_data[0, 0] = self.set_fwrap_metadata()
        fwrap_data[1, 0] = MatlabCanonicalEmpty()

        for i, prop_val in enumerate(self.mcos_props_saved):
            fwrap_data[2 + i, 0] = prop_val

        # Write Unknowns
        u3_arr = np.empty((self.class_id_counter + 1, 1), dtype=object)
        for i in range(self.class_id_counter + 1):
            u3_arr[i, 0] = EmptyMatStruct(np.empty((1, 0), dtype=object))
        fwrap_data[-3, 0] = u3_arr
        fwrap_data[-1, 0] = u3_arr

        fwrap_data[-2, 0] = np.array(
            self.mcos_class_alias_metadata, dtype=np.uint32
        ).reshape(-1, 1)

        fwrapper = MatlabOpaque(
            properties=fwrap_data,
            classname=MCOS_SUBSYSTEM_CLASS,
            type_system=OpaqueType.MCOS,
        )
        return fwrapper

    def set_subsystem(self):
        """Creates subsystem struct array"""

        fwrap_data = self.set_fwrap_data()
        if fwrap_data is None:
            return None

        dtype = [(OpaqueType.MCOS, object)]
        subsys_metadata = np.empty((1, 1), dtype=dtype)
        subsys_metadata[0, 0][OpaqueType.MCOS] = fwrap_data

        return subsys_metadata

    def load_mcos_enumeration(self, metadata, type_system):
        """Loads MATLAB MCOS enumeration instance array"""

        classname = self.get_classname(metadata[0, 0]["ClassName"].item())
        builtin_class_idx = metadata[0, 0]["BuiltinClassName"].item()
        if builtin_class_idx != 0:
            builtin_class_name = self.get_classname(builtin_class_idx)
        else:
            builtin_class_name = np.str_("")

        value_names = [
            self.mcos_names[val - 1] for val in metadata[0, 0]["ValueNames"].ravel()
        ]

        enum_vals = []
        value_idx = metadata[0, 0]["ValueIndices"]
        mmdata = metadata[0, 0]["Values"]  # Array is N x 1 shape
        if mmdata.size != 0:
            mmdata_map = mmdata[value_idx]
            for val in np.nditer(mmdata_map, flags=["refs_ok"], op_flags=["readonly"]):
                obj_array = self.load_mcos_object(val.item(), "MCOS")
                enum_vals.append(obj_array)

        if not self.raw_data:
            return mat_to_enum(
                enum_vals,
                value_names,
                classname,
                value_idx.shape,
            )

        metadata[0, 0]["BuiltinClassName"] = builtin_class_name
        metadata[0, 0]["ClassName"] = classname
        metadata[0, 0]["ValueNames"] = np.array(value_names).reshape(
            value_idx.shape, order="F"
        )
        metadata[0, 0]["ValueIndices"] = value_idx
        metadata[0, 0]["Values"] = np.array(enum_vals).reshape(
            value_idx.shape, order="F"
        )

        return MatlabOpaque(
            metadata,
            classname,
            type_system,
            self.get_class_alias(metadata[0, 0]["ClassName"].item()),
        )

    def load_mcos_object(self, metadata, type_system=OpaqueType.MCOS):
        """Loads MCOS object"""

        metadata = np.atleast_2d(metadata)

        ndims = metadata[1, 0]
        dims = metadata[2 : 2 + ndims, 0]
        nobjects = np.prod(dims)
        object_ids = metadata[2 + ndims : 2 + ndims + nobjects, 0]

        class_id = metadata[-1, 0]
        classname = self.get_classname(class_id)
        class_alias = self.get_class_alias(class_id)

        if object_ids.size == 0:
            return MatlabOpaque(
                properties=tuple(dims),
                classname=classname,
                type_system=type_system,
                class_alias=class_alias,
            )

        is_array = nobjects > 1
        array_objs = []
        for object_id in object_ids:
            if object_id in self.mcos_object_cache:
                obj = self.mcos_object_cache[object_id]
            else:
                if not self.raw_data and classname in matlab_classdef_types:
                    obj_props = self.get_properties(object_id)
                    obj = convert_mat_to_py(
                        obj_props,
                        classname,
                        byte_order=self.byte_order,
                        add_table_attrs=self.add_table_attrs,
                    )
                    self.mcos_object_cache[object_id] = (
                        obj  # Caching here is probably unnecessary but safer
                    )
                else:
                    obj = MatlabOpaque(
                        properties=None, classname=classname, type_system=type_system
                    )
                    self.mcos_object_cache[object_id] = obj
                    obj.class_alias = class_alias
                    obj.properties = self.get_properties(object_id)
            array_objs.append(obj)

        if is_array:
            obj_arr = np.empty((nobjects,), dtype=object)
            obj_arr[:] = array_objs
            obj_arr = obj_arr.reshape(dims, order="F")
            obj_arr = MatlabOpaqueArray(obj_arr, classname, type_system)
        else:
            obj_arr = array_objs[0]

        return obj_arr

    def load_opaque_object(self, metadata, type_system, classname=None):
        """Loads opaque object"""

        if type_system != OpaqueType.MCOS:
            warnings.warn(
                f"Opaque object of type {type_system} is not supported",
                MatReadWarning,
                stacklevel=2,
            )
            return MatlabOpaque(metadata, classname, type_system)

        if metadata.dtype.names:
            return self.load_mcos_enumeration(metadata, type_system)
        else:
            return self.load_mcos_object(metadata, type_system)



"""Utility functions for converting MATLAB tables and timetables to pandas DataFrames"""

import warnings

import numpy as np
import pandas as pd

from matio.utils.converters.mattimes import caldur_dtype
from matio.utils.matclass import EmptyMatStruct, MatConvertError, MatConvertWarning

TABLE_VERSION = 4
MIN_TABLE_VERSION = 1

TIMETABLE_VERSION = 6
MIN_TIMETABLE_VERSION = 2

# Pandas marks this as experimental
# Using it here as its the closest match to MATLAB tables
# FIXME: Update with pandas 3.0 release
# https://pandas.pydata.org/docs/user_guide/migration-3-strings.html
pd.options.future.infer_string = True


def add_table_props(df, tab_props):
    """Add MATLAB table properties to pandas DataFrame
    These properties are mostly cell arrays of character vectors
    """

    tab_props = tab_props[0, 0]

    df.attrs["Description"] = (
        tab_props["Description"].item() if tab_props["Description"].size > 0 else ""
    )
    df.attrs["VariableDescriptions"] = [
        s.item() if s.size > 0 else ""
        for s in tab_props["VariableDescriptions"].ravel()
    ]
    df.attrs["VariableUnits"] = [
        s.item() if s.size > 0 else "" for s in tab_props["VariableUnits"].ravel()
    ]
    df.attrs["VariableContinuity"] = [
        s.item() if s.size > 0 else "" for s in tab_props["VariableContinuity"].ravel()
    ]
    df.attrs["DimensionNames"] = [
        s.item() if s.size > 0 else "" for s in tab_props["DimensionNames"].ravel()
    ]
    df.attrs["UserData"] = tab_props["UserData"]

    return df


def add_timetable_props(df, tab_props):
    """Add MATLAB table properties to pandas DataFrame
    These properties are mostly cell arrays of character vectors
    """
    df.attrs["varDescriptions"] = [
        s.item() if s.size > 0 else "" for s in tab_props["varDescriptions"].ravel()
    ]
    df.attrs["varUnits"] = [
        s.item() if s.size > 0 else "" for s in tab_props["varUnits"].ravel()
    ]
    df.attrs["varContinuity"] = [
        s.item() if s.size > 0 else "" for s in tab_props["varContinuity"].ravel()
    ]
    df.attrs["UserData"] = tab_props["arrayProps"]["UserData"][0, 0]
    df.attrs["Description"] = (
        tab_props["arrayProps"]["Description"][0, 0].item()
        if tab_props["arrayProps"]["Description"][0, 0].size > 0
        else ""
    )

    return df


def to_dataframe(data, nvars, varnames):
    """Creates a dataframe from coldata and column names"""

    def make_series(col, dtype_kind):
        if dtype_kind == "T":
            return pd.Series(col, dtype="str")
        else:
            return pd.Series(col)

    rows = {}
    for i in range(nvars):
        vname = varnames[0, i].item()
        coldata = data[0, i]
        if isinstance(coldata, np.ndarray):
            if coldata.shape[1] == 1:
                rows[vname] = make_series(coldata[:, 0], coldata.dtype.kind)
            else:
                for j in range(coldata.shape[1]):
                    colname = f"{vname}_{j + 1}"
                    rows[colname] = make_series(coldata[:, j], coldata.dtype.kind)
        else:
            rows[vname] = pd.Series(coldata)

    df = pd.DataFrame(rows)
    return df


def mat_to_table(props, add_table_attrs=False, **_kwargs):
    """Converts MATLAB table to pandas DataFrame"""

    table_attrs = props.get("props")
    ver = int(table_attrs[0, 0]["versionSavedFrom"].item())
    if ver > TABLE_VERSION:
        warnings.warn(
            f"mat_to_table: MATLAB table version {ver} is not supported.",
            UserWarning,
        )
        return props

    data = props.get("data")
    nvars = int(props.get("nvars").item())
    varnames = props.get("varnames")
    df = to_dataframe(data, nvars, varnames)

    # Add df.index
    nrows = int(props.get("nrows").item())
    rownames = props.get("rownames")
    if rownames.size > 0:
        rownames = [s.item() for s in rownames.ravel()]
        if len(rownames) == nrows:
            df.index = rownames

    if add_table_attrs:
        # Since pandas lists this as experimental, flag so we can switch off if it breaks
        df = add_table_props(df, table_attrs)

    return df


def get_row_times(row_times, num_rows):
    """Get row times from MATLAB timetable
    rowTimes is a duration or datetime array if explicitly specified
    If using "SampleRate" or "TimeStep", it is a struct array with the following fields:
    1. origin - the start time as a duration or datetime scalar
    2. specifiedAsRate - boolean indicating which to use - sampleRate or TimeStep
    3. stepSize - the time step as a duration scalar
    4. sampleRate - the sample rate as a float
    """
    if not row_times.dtype.names:
        return row_times.ravel()

    start = row_times[0, 0]["origin"]
    if row_times[0, 0]["specifiedAsRate"]:
        fs = row_times[0, 0]["sampleRate"].item()
        step = np.timedelta64(int(1e9 / fs), "ns")
        warnings.warn(
            "get_row_times: MATLAB sampleRate is converted to TimeStep in nanoseconds for NumPy compatibility.",
            MatConvertWarning,
        )
    else:
        step = None
        comps = row_times[0, 0]["stepSize"]

        if comps.dtype == caldur_dtype:
            # calendarDuration
            # Only one of months, days, or millis is non-zero array
            for field in comps.dtype.names:
                if np.any(comps[field] != 0):
                    step = comps[field]
                    break
            if step is None:
                step = comps[0, 0]["millis"]  # fallback if all are zero

            step_unit = np.datetime_data(step.dtype)[0]
            start = start.astype(f"datetime64[{step_unit}]")
        else:
            step = comps
            step_dtype_unit = np.datetime_data(step.dtype)[0]

            if start.dtype.kind == "m":
                start_dtype_new = f"timedelta64[{step_dtype_unit}]"
            else:
                start_dtype_new = f"datetime64[{step_dtype_unit}]"
            start = start.astype(start_dtype_new)

    times = (start + step * np.arange(num_rows)).ravel()
    if times.dtype.kind == "M":
        # MATLAB saves datetime as [ms] data
        times = times.astype("datetime64[ns]")
    return times


def mat_to_timetable(props, add_table_attrs=False, **_kwargs):
    """Converts MATLAB timetable to pandas DataFrame"""

    timetable_data = props.get("any", None)
    if timetable_data is None:
        return props

    ver = int(timetable_data[0, 0]["versionSavedFrom"].item())
    if ver > TIMETABLE_VERSION or ver <= MIN_TIMETABLE_VERSION:
        warnings.warn(
            f"mat_to_timetable: MATLAB timetable version {ver} is not supported.",
            UserWarning,
        )
        return props

    num_vars = int(timetable_data[0, 0]["numVars"].item())
    var_names = timetable_data[0, 0]["varNames"]
    data = timetable_data[0, 0]["data"]
    df = to_dataframe(data, num_vars, var_names)

    row_times = timetable_data[0, 0]["rowTimes"]
    num_rows = int(timetable_data[0, 0]["numRows"].item())

    row_times = get_row_times(row_times, num_rows)
    dim_names = timetable_data[0, 0]["dimNames"]
    df.index = pd.Index(row_times, name=dim_names[0, 0].item())

    if add_table_attrs:
        # Since pandas lists this as experimental, flag so we can switch off if it breaks
        df = add_timetable_props(df, timetable_data[0, 0])

    return df


def mat_to_categorical(props, **_kwargs):
    """Converts MATLAB categorical to pandas Categorical
    MATLAB categorical objects are stored with the following properties:
    1. categoryNames - all unique categories
    2. codes
    3. isOrdinal - boolean indicating if the categorical is ordered
    4. isProtected - boolean indicating if the categorical is protected
    """

    raw_names = props.get("categoryNames")
    category_names = [name.item() for name in raw_names.ravel()]

    # MATLAB codes are 1-indexed as uint integers
    codes = props.get("codes").astype(int) - 1
    ordered = bool(props.get("isOrdinal").item())
    return pd.Categorical.from_codes(codes, categories=category_names, ordered=ordered)


def make_table_props():
    """Creates default properties for a MATLAB table"""
    dtype = [
        ("useVariableNamesOriginal", object),
        ("useDimensionNamesOriginal", object),
        ("CustomProps", object),
        ("VariableCustomProps", object),
        ("versionSavedFrom", object),
        ("minCompatibleVersion", object),
        ("incompatibilityMsg", object),
        ("VersionSavedFrom", object),
        ("Description", object),
        ("VariableNamesOriginal", object),
        ("DimensionNames", object),
        ("DimensionNamesOriginal", object),
        ("UserData", object),
        ("VariableDescriptions", object),
        ("VariableUnits", object),
        ("VariableContinuity", object),
    ]

    props = np.empty((1, 1), dtype=dtype)

    props["useVariableNamesOriginal"][0, 0] = np.bool_(False)
    props["useDimensionNamesOriginal"][0, 0] = np.bool_(False)
    props["CustomProps"][0, 0] = EmptyMatStruct(np.empty((1, 1), dtype=object))
    props["VariableCustomProps"][0, 0] = EmptyMatStruct(np.empty((1, 1), dtype=object))
    props["versionSavedFrom"][0, 0] = np.float64(TABLE_VERSION)
    props["minCompatibleVersion"][0, 0] = np.float64(MIN_TABLE_VERSION)
    props["incompatibilityMsg"][0, 0] = np.empty((0, 0), dtype=np.str_)
    props["VersionSavedFrom"][0, 0] = np.float64(TABLE_VERSION)
    props["Description"][0, 0] = np.empty((0, 0), dtype=np.str_)
    props["VariableNamesOriginal"][0, 0] = np.empty((0, 0), dtype=object)
    props["DimensionNames"][0, 0] = np.array(
        [np.array(["Row"]), np.array(["Variables"])], dtype=object
    ).reshape((1, 2))
    props["DimensionNamesOriginal"][0, 0] = np.empty((0, 0), dtype=object)
    props["UserData"][0, 0] = np.empty((0, 0), dtype=np.float64)
    props["VariableDescriptions"][0, 0] = np.empty((0, 0), dtype=object)
    props["VariableUnits"][0, 0] = np.empty((0, 0), dtype=object)
    props["VariableContinuity"][0, 0] = np.empty((0, 0), dtype=object)

    return props


def table_to_mat(df):
    """Converts a pandas DataFrame to a MATLAB table"""

    data = np.empty((1, len(df.columns)), dtype=object)
    for i, col in enumerate(df.columns):
        if df[col].dtype == "str":
            coldata = (
                df[col]
                .to_numpy(dtype=np.dtypes.StringDType(na_object=np.nan))
                .reshape(-1, 1)
            )
        else:
            coldata = df[col].to_numpy().reshape(-1, 1)
        data[0, i] = coldata

    nrows = np.float64(df.shape[0])
    nvars = np.float64(df.shape[1])

    varnames = np.array([str(col) for col in df.columns], dtype=object)

    if df.index.name is not None or not isinstance(df.index, pd.RangeIndex):
        rownames = np.array([str(idx) for idx in df.index], dtype=object)
    else:
        rownames = np.empty((0, 0), dtype=object)

    # All serialized data must be the exact same types (including shape) as expected by MATLAB
    # If not, it might still load correctly, but some class methods may fail
    extras = make_table_props()
    prop_map = {
        "data": data,
        "varnames": varnames,
        "nrows": nrows,
        "nvars": nvars,
        "rownames": rownames,
        "ndims": np.float64(2),
        "props": extras,
    }

    return prop_map


def make_timetable_props():
    """Creates default properties for a MATLAB timetable"""

    arrayprops_dtype = [
        ("Description", object),
        ("UserData", object),
        ("TableCustomProperties", object),
    ]
    arrayprops = np.empty((1, 1), dtype=arrayprops_dtype)
    arrayprops["Description"][0, 0] = np.empty((0, 0), dtype=np.str_)
    arrayprops["UserData"][0, 0] = np.empty((0, 0), dtype=np.float64)
    arrayprops["TableCustomProperties"][0, 0] = EmptyMatStruct(
        np.empty((1, 1), dtype=object)
    )

    return {
        "CustomProps": EmptyMatStruct(np.empty((1, 1), dtype=object)),
        "VariableCustomProps": EmptyMatStruct(np.empty((1, 1), dtype=object)),
        "versionSavedFrom": np.float64(TIMETABLE_VERSION),
        "minCompatibleVersion": np.float64(MIN_TIMETABLE_VERSION),
        "incompatibilityMsg": np.empty((0, 0), dtype=np.str_),
        "arrayProps": arrayprops,
        "numDims": np.float64(2),
        "useVarNamesOrig": np.bool_(False),
        "useDimNamesOrig": np.bool_(False),
        "dimNamesOrig": np.empty((0, 0), dtype=object),
        "varNamesOrig": np.empty((0, 0), dtype=object),
        "varDescriptions": np.empty((0, 0), dtype=object),
        "varUnits": np.empty((0, 0), dtype=object),
        "timeEvents": np.empty((0, 0), dtype=np.float64),
        "varContinuity": np.empty((0, 0), dtype=object),
    }


def timetable_to_mat(df):
    """Converts a pandas DataFrame to a MATLAB timetable"""

    data = np.empty((1, len(df.columns)), dtype=object)
    for i, col in enumerate(df.columns):
        if df[col].dtype == "str":
            coldata = (
                df[col]
                .to_numpy(dtype=np.dtypes.StringDType(na_object=np.nan))
                .reshape(-1, 1)
            )
        else:
            coldata = df[col].to_numpy().reshape(-1, 1)
        data[0, i] = coldata

    nrows = np.float64(df.shape[0])
    nvars = np.float64(df.shape[1])

    varnames = np.array([str(col) for col in df.columns], dtype=object)

    dim1 = df.index.name if df.index.name is not None else "Time"
    dimnames = np.array(
        [np.array([dim1]), np.array(["Variables"])], dtype=object
    ).reshape((1, 2))

    if isinstance(df.index, (pd.DatetimeIndex, pd.TimedeltaIndex)):
        rowtimes = df.index.to_numpy()
    else:
        raise MatConvertError(
            "timetable_to_mat: DataFrame index must be DatetimeIndex or TimedeltaIndex."
        )

    # Define timetable struct dtype
    timetable_dtype = [
        ("data", object),
        ("dimNames", object),
        ("varNames", object),
        ("numRows", object),
        ("numVars", object),
        ("rowTimes", object),
    ]

    extras = make_timetable_props()
    timetable_dtype.extend((key, object) for key in extras)

    # Create 1x1 structured array
    timetable = np.empty((1, 1), dtype=timetable_dtype)
    timetable[0, 0]["data"] = data
    timetable[0, 0]["dimNames"] = dimnames
    timetable[0, 0]["varNames"] = varnames.reshape((1, -1))
    timetable[0, 0]["numRows"] = nrows
    timetable[0, 0]["numVars"] = nvars
    timetable[0, 0]["rowTimes"] = rowtimes.reshape((-1, 1))

    for key, value in extras.items():
        timetable[0, 0][key] = value

    return {"any": timetable}


def categorical_to_mat(cat):
    """Converts a pandas Categorical to a MATLAB categorical"""

    category_names = cat.categories.to_numpy(dtype=object).reshape(-1, 1)
    codes = cat.codes.astype("int8") + 1  # 1-based indexing
    is_ordinal = np.bool_(cat.ordered)
    is_protected = np.bool_(False)  # not supported in pandas

    return {
        "categoryNames": category_names,
        "codes": codes,
        "isOrdinal": is_ordinal,
        "isProtected": is_protected,
    }



import os
import tempfile

import numpy as np
import pytest

from matio import load_from_mat, save_to_mat
from matio.utils.matclass import MatlabContainerMap

files = [("test_maps_v7.mat", "v7"), ("test_maps_v73.mat", "v7.3")]
namespace = "TestClasses"


@pytest.mark.parametrize("filename, version", files)
class TestLoadMatlabContainerMap:

    def test_map_char_keys(self, filename, version):
        """Test reading container.Map with char keys from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["map_char_keys"])
        assert "map_char_keys" in mdict
        assert isinstance(mdict["map_char_keys"], MatlabContainerMap)

        d = {
            "a": np.array([[1]], dtype=np.float64),
            "b": np.array([[2]], dtype=np.float64),
        }

        for key, val in d.items():
            assert key in mdict["map_char_keys"]
            if isinstance(val, np.ndarray):
                np.testing.assert_array_equal(
                    mdict["map_char_keys"][key], val, strict=True
                )
            else:
                assert mdict["map_char_keys"][key] == val

    def test_map_numeric_keys(self, filename, version):
        """Test reading container.Map with numeric keys from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["map_numeric_keys"])
        assert "map_numeric_keys" in mdict
        assert isinstance(mdict["map_numeric_keys"], MatlabContainerMap)

        d = {
            1: np.array(["a"]),
            2: np.array(["b"]),
        }

        for key, val in d.items():
            assert key in mdict["map_numeric_keys"]
            if isinstance(val, np.ndarray):
                np.testing.assert_array_equal(
                    mdict["map_numeric_keys"][key], val, strict=True
                )
            else:
                assert mdict["map_numeric_keys"][key] == val

    def test_map_string_keys(self, filename, version):
        """Test reading container.Map with char keys (from string) from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["map_string_keys"])
        assert "map_string_keys" in mdict
        assert isinstance(mdict["map_string_keys"], MatlabContainerMap)

        d = {
            "a": np.array([[1]], dtype=np.float64),
            "b": np.array([[2]], dtype=np.float64),
        }

        for key, val in d.items():
            assert key in mdict["map_string_keys"]
            if isinstance(val, np.ndarray):
                np.testing.assert_array_equal(
                    mdict["map_string_keys"][key], val, strict=True
                )
            else:
                assert mdict["map_string_keys"][key] == val

    def test_map_empty(self, filename, version):
        """Test reading container.Map empty from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["map_empty"])
        assert "map_empty" in mdict
        assert isinstance(mdict["map_empty"], MatlabContainerMap)
        assert len(mdict["map_empty"]) == 0


@pytest.mark.parametrize("filename, version", files)
class TestSaveMatlabContainerMap:

    def test_map_char_keys(self, filename, version):
        """Test writing container.Map with char keys to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["map_char_keys"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["map_char_keys"])

            for key, val in mdict["map_char_keys"].items():
                assert key in mload["map_char_keys"]
                if isinstance(val, np.ndarray):
                    np.testing.assert_array_equal(
                        mload["map_char_keys"][key], val, strict=True
                    )
                else:
                    assert mload["map_char_keys"][key] == val

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_map_numeric_keys(self, filename, version):
        """Test writing container.Map with numeric keys to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["map_numeric_keys"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["map_numeric_keys"])

            for key, val in mdict["map_numeric_keys"].items():
                assert key in mload["map_numeric_keys"]
                if isinstance(val, np.ndarray):
                    np.testing.assert_array_equal(
                        mload["map_numeric_keys"][key], val, strict=True
                    )
                else:
                    assert mload["map_numeric_keys"][key] == val

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_map_string_keys(self, filename, version):
        """Test writing container.Map with char keys to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["map_string_keys"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["map_string_keys"])

            for key, val in mdict["map_string_keys"].items():
                assert key in mload["map_string_keys"]
                if isinstance(val, np.ndarray):
                    np.testing.assert_array_equal(
                        mload["map_string_keys"][key], val, strict=True
                    )
                else:
                    assert mload["map_string_keys"][key] == val

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_map_empty(self, filename, version):
        """Test writing container.Map empty to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["map_empty"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["map_empty"])
            assert len(mload["map_empty"]) == 0

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)



import os
import tempfile

import numpy as np
import pytest

from matio import load_from_mat, save_to_mat
from matio.utils.matclass import MatlabOpaque, MatlabOpaqueProperty

files = [("test_dynamic_v7.mat", "v7"), ("test_dynamic_v73.mat", "v7.3")]
namespace = "TestClasses"


@pytest.mark.parametrize("filename, version", files)
class TestLoadMatlabDynamic:

    def test_dynamicprop(self, filename, version):
        """Test reading dynamic property from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["obj"])

        assert "obj" in mdict
        assert isinstance(mdict["obj"], MatlabOpaque)
        assert mdict["obj"].classname == f"{namespace}.BasicDynamic"
        assert mdict["obj"].type_system == "MCOS"

        assert "__dynamic_property__1" in mdict["obj"].properties
        assert isinstance(
            mdict["obj"].properties["__dynamic_property__1"], MatlabOpaque
        )
        assert (
            mdict["obj"].properties["__dynamic_property__1"].classname
            == "meta.DynamicProperty"
        )
        assert mdict["obj"].properties["__dynamic_property__1"].type_system == "MCOS"


@pytest.mark.parametrize("filename, version", files)
class TestSaveMatlabDynamic:

    def test_dynamicprops(self, filename, version):
        """Test writing dynamic properties to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["obj"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["obj"])

            assert "obj" in mload
            assert isinstance(mload["obj"], MatlabOpaque)
            assert mload["obj"].classname == mdict["obj"].classname
            assert mload["obj"].type_system == mdict["obj"].type_system

            assert "__dynamic_property__1" in mload["obj"].properties
            assert isinstance(
                mload["obj"].properties["__dynamic_property__1"], MatlabOpaque
            )
            assert (
                mload["obj"].properties["__dynamic_property__1"].classname
                == mdict["obj"].properties["__dynamic_property__1"].classname
            )
            assert (
                mload["obj"].properties["__dynamic_property__1"].type_system
                == mdict["obj"].properties["__dynamic_property__1"].type_system
            )

            for key, val in mdict["obj"].properties.items():
                if isinstance(val, MatlabOpaque):
                    assert key in mload["obj"].properties
                    assert isinstance(mload["obj"].properties[key], MatlabOpaque)
                    assert mload["obj"].properties[key].classname == val.classname
                    assert mload["obj"].properties[key].type_system == val.type_system
                elif isinstance(val, MatlabOpaqueProperty):
                    assert key in mload["obj"].properties
                    assert mload["obj"].properties[key].ptype == val.ptype
                    np.testing.assert_array_equal(
                        mload["obj"].properties[key], val, strict=True
                    )
                elif isinstance(val, np.ndarray):
                    assert key in mload["obj"].properties
                    np.testing.assert_array_equal(
                        mload["obj"].properties[key], val, strict=True
                    )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)



"""Utility functions for converting MATLAB datetime, duration, and calendarDuration"""

import warnings

import numpy as np

from matio.utils.matclass import MatConvertWarning

caldur_dtype = [
    ("months", "timedelta64[M]"),
    ("days", "timedelta64[D]"),
    ("millis", "timedelta64[ms]"),
]


def mat_to_datetime(props, **_kwargs):
    """Convert MATLAB datetime to Numpy datetime64 array"""

    data = props.get("data", None)
    if data is None or data.size == 0:
        return np.empty((0, 0), dtype="datetime64[ns]")

    tz = props.get("tz", None)
    if tz is not None and tz.size > 0:
        warnings.warn(
            f"MATLAB datetime was saved with timezone '{tz}'."
            "NumPy datetime64 does not support time zones, so the timezone information will be lost.",
            MatConvertWarning,
            stacklevel=2,
        )

    fmt = props.get("fmt", None)
    if fmt is not None and fmt.size > 0:
        warnings.warn(
            "mat_to_datetime: Ignoring 'fmt' property. Numpy datetime64 does not support custom formats.",
            MatConvertWarning,
            stacklevel=2,
        )

    # * Does MATLAB handle sub-ms precision?
    # I seem to recall that MATLAB encodes sub-ms precision in the imaginary part of complex numbers.
    # But I can't seem to replicate this, maybe it was specific to a certain version?

    ms_frac, ms_int_float = np.modf(data)
    ns_int = ms_int_float.astype(np.int64) * 1000_000
    ns_frac = np.round(ms_frac * 1e6).astype(np.int64)
    total_ns = ns_int + ns_frac
    return total_ns.astype("datetime64[ns]")


def mat_to_duration(props, **_kwargs):
    """Convert MATLAB duration to Numpy timedelta64 array"""

    millis = props["millis"]
    # if millis.size == 0:
    # return np.empty((0, 0), dtype="timedelta64[ms]")

    fmt = props.get("fmt", None)
    if fmt is None:
        return millis.astype("timedelta64[ms]")

    if fmt == "s":
        count = millis / 1000  # Seconds
        dur = count.astype("timedelta64[s]")
    elif fmt == "m":
        count = millis / (1000 * 60)  # Minutes
        dur = count.astype("timedelta64[m]")
    elif fmt == "h":
        count = millis / (1000 * 60 * 60)  # Hours
        dur = count.astype("timedelta64[h]")
    elif fmt == "d":
        count = millis / (1000 * 60 * 60 * 24)  # Days
        dur = count.astype("timedelta64[D]")
    elif fmt == "y":
        count = millis / (1000 * 60 * 60 * 24 * 365)  # Years
        dur = count.astype("timedelta64[Y]")
    else:
        count = millis
        dur = count.astype("timedelta64[ms]")
        warnings.warn(
            f"mat_to_duration: Unknown format '{fmt}'. Defaulting to 'ms'.",
            MatConvertWarning,
            stacklevel=2,
        )
        # Default case

    return dur


def mat_caldur_to_numpy(arr):
    """Converts MATLAB caledarDuration struct to numpy timedelta record"""

    def _broadcast_to_ref(x, dtype):
        x = np.asarray(x)
        if x.size == 1:
            return np.full(ref_shape, x.item(), dtype=dtype)
        return x.astype(dtype)

    months = arr[0, 0]["months"]
    days = arr[0, 0]["days"]
    millis = arr[0, 0]["millis"]

    ref_shape = max([x.shape for x in (months, days, millis)], key=lambda s: np.prod(s))
    res = np.empty(ref_shape, dtype=caldur_dtype)
    res["months"] = _broadcast_to_ref(months, "timedelta64[M]")
    res["days"] = _broadcast_to_ref(days, "timedelta64[D]")
    res["millis"] = _broadcast_to_ref(millis, "timedelta64[ms]")

    return res


def mat_to_calendarduration(props, **_kwargs):
    """Convert MATLAB calendarDuration to Dict of Python Timedeltas"""

    comps = props.get("components", None)
    if comps is None:
        return props

    arr = mat_caldur_to_numpy(comps)
    return arr


def datetime_to_mat(arr):
    """Convert numpy.datetime64 array to MATLAB datetime format."""

    unit = np.datetime_data(arr.dtype)[0]

    # For sub-millisecond precision, preserve in decimal portion
    if unit in ["us", "ns", "ps", "fs", "as"]:
        dt_int = arr.astype(np.int64)

        unit_to_ms = {"us": 1e-3, "ns": 1e-6, "ps": 1e-9, "fs": 1e-12, "as": 1e-15}

        conversion_factor = unit_to_ms[unit]
        millis = dt_int.astype(np.float64) * conversion_factor
    else:
        millis = arr.astype("datetime64[ms]").astype(np.int64).astype(np.float64)

    tz = np.empty((0, 0), dtype=np.str_)
    fmt = np.empty((0, 0), dtype=np.str_)
    prop_map = {"data": millis, "tz": tz, "fmt": fmt}

    return prop_map


def duration_to_mat(arr):
    """Convert numpy timedelta64 array to MATLAB duration format."""

    unit, _ = np.datetime_data(arr.dtype)
    millis = arr.astype("timedelta64[ns]").astype(np.float64) / 1e6
    allowed_units = ("s", "m", "h", "D", "Y")
    if unit not in allowed_units:
        warnings.warn(
            f"duration_to_mat: MATLAB Duration arrays do not support timedelta64[{unit}]. Defaulting to 's'.",
            MatConvertWarning,
        )
        unit = "s"

    unit = unit.lower()
    prop_map = {
        "millis": millis,
        "fmt": unit,
    }

    return prop_map


def numpy_to_matcaldur(arr):
    """Convert numpy structured array with fields ['months', 'days', 'millis'] to MATLAB calendarDuration format."""

    fields = ["months", "days", "millis"]
    comp_dtype = [(f, object) for f in fields]
    comps = np.empty((1, 1), dtype=comp_dtype)

    if arr.size == 0:
        for f in fields:
            comps[0, 0][f] = np.empty((0, 0), dtype=np.float64)
        return comps

    for f in fields:
        if np.all(arr[f] == arr[f].flat[0]):
            # MATLAB reduces to scalar if all values are the same
            reduced = np.array([[arr[f].flat[0]]], dtype=arr[f].dtype)
            comps[0, 0][f] = reduced.astype(np.float64)
        else:
            comps[0, 0][f] = arr[f].astype(np.float64)

    return comps


def calendarduration_to_mat(arr):
    """Convert numpy structured array with fields ['months', 'days', 'millis'] to MATLAB calendarDuration format."""

    comps = numpy_to_matcaldur(arr)
    fmt = "ymwdt"

    prop_map = {
        "components": comps,
        "fmt": fmt,
    }

    return prop_map



import os
import tempfile

import numpy as np
import pandas as pd
import pytest

from matio import load_from_mat, save_to_mat
from matio.utils.matclass import MatlabOpaque

files = [("test_tables_v7.mat", "v7"), ("test_tables_v73.mat", "v7.3")]
namespace = "TestClasses"


@pytest.mark.parametrize("filename, version", files)
class TestLoadMatlabCategorical:

    def test_categorical_scalar(self, filename, version):
        """Test reading categorical scalar data from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cat_scalar"])
        assert "cat_scalar" in mdict

        cats = np.array(["blue", "green", "red"], dtype=object)
        codes = np.array([[2, 1, 0, 2]], dtype=np.int8)
        ordered = False
        np.testing.assert_array_equal(mdict["cat_scalar"].codes, codes, strict=True)
        np.testing.assert_array_equal(mdict["cat_scalar"].categories, cats, strict=True)
        np.testing.assert_array_equal(mdict["cat_scalar"].ordered, ordered, strict=True)

    def test_categorical_array(self, filename, version):
        """Test reading categorical array data from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cat_array"])
        assert "cat_array" in mdict

        cats = np.array(["high", "low", "medium"], dtype=object)
        codes = np.array([[1, 2], [0, 1]], dtype=np.int8)
        ordered = False

        np.testing.assert_array_equal(mdict["cat_array"].codes, codes, strict=True)
        np.testing.assert_array_equal(mdict["cat_array"].categories, cats, strict=True)
        np.testing.assert_array_equal(mdict["cat_array"].ordered, ordered, strict=True)

    def test_categorical_empty(self, filename, version):
        """Test reading empty categorical data from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cat_empty"])
        assert "cat_empty" in mdict

        cats = np.array([], dtype=object)
        codes = np.empty((0, 0), dtype=np.int8)
        ordered = False

        np.testing.assert_array_equal(mdict["cat_empty"].codes, codes, strict=True)
        np.testing.assert_array_equal(mdict["cat_empty"].categories, cats, strict=True)
        np.testing.assert_array_equal(mdict["cat_empty"].ordered, ordered, strict=True)

    def test_categorical_from_numeric(self, filename, version):
        """Test reading categorical data created from numeric array from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cat_from_numeric"])
        assert "cat_from_numeric" in mdict

        cats = np.array(["low", "medium", "high"], dtype=object)
        codes = np.array([[0, 1, 2, 1, 0]], dtype=np.int8)
        ordered = False

        np.testing.assert_array_equal(
            mdict["cat_from_numeric"].codes, codes, strict=True
        )
        np.testing.assert_array_equal(
            mdict["cat_from_numeric"].categories, cats, strict=True
        )
        np.testing.assert_array_equal(
            mdict["cat_from_numeric"].ordered, ordered, strict=True
        )

    def test_categorical_unordered(self, filename, version):
        """Test reading unordered categorical data from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cat_unordered"])
        assert "cat_unordered" in mdict

        cats = np.array(["cold", "warm", "hot"], dtype=object)
        codes = np.array([[0, 2, 1]], dtype=np.int8)
        ordered = False

        np.testing.assert_array_equal(mdict["cat_unordered"].codes, codes, strict=True)
        np.testing.assert_array_equal(
            mdict["cat_unordered"].categories, cats, strict=True
        )
        np.testing.assert_array_equal(
            mdict["cat_unordered"].ordered, ordered, strict=True
        )

    def test_categorical_ordered(self, filename, version):
        """Test reading ordered categorical data from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cat_ordered"])
        assert "cat_ordered" in mdict

        cats = np.array(["small", "medium", "large"], dtype=object)
        codes = np.array([[0, 1, 2]], dtype=np.int8)
        ordered = True

        np.testing.assert_array_equal(mdict["cat_ordered"].codes, codes, strict=True)
        np.testing.assert_array_equal(
            mdict["cat_ordered"].categories, cats, strict=True
        )
        np.testing.assert_array_equal(
            mdict["cat_ordered"].ordered, ordered, strict=True
        )

    def test_categorical_missing(self, filename, version):
        """Test reading categorical data with missing values from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cat_missing"])
        assert "cat_missing" in mdict

        cats = np.array(["cat", "dog", "mouse"], dtype=object)
        codes = np.array([[0, -1, 1, 2]], dtype=np.int8)
        ordered = False

        np.testing.assert_array_equal(mdict["cat_missing"].codes, codes, strict=True)
        np.testing.assert_array_equal(
            mdict["cat_missing"].categories, cats, strict=True
        )
        np.testing.assert_array_equal(
            mdict["cat_missing"].ordered, ordered, strict=True
        )

    def test_categorical_mixed_case(self, filename, version):
        """Test reading categorical data with mixed case categories from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cat_mixed_case"])
        assert "cat_mixed_case" in mdict

        cats_list = sorted(["On", "off", "OFF", "ON", "on"])
        cats = np.array(cats_list, dtype=object)
        codes = np.array([[2, 3, 0, 1, 4]], dtype=np.int8)
        ordered = False

        np.testing.assert_array_equal(mdict["cat_mixed_case"].codes, codes, strict=True)
        np.testing.assert_array_equal(
            mdict["cat_mixed_case"].categories, cats, strict=True
        )
        np.testing.assert_array_equal(
            mdict["cat_mixed_case"].ordered, ordered, strict=True
        )

    def test_categorical_matlab_string(self, filename, version):
        """Test reading categorical data with categories as MATLAB strings from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cat_string"])
        assert "cat_string" in mdict

        cats_list = sorted(["spring", "summer", "autumn", "winter"])
        cats = np.array(cats_list, dtype=object)
        codes = np.array([[1, 2, 0, 3]], dtype=np.int8)
        ordered = False

        np.testing.assert_array_equal(mdict["cat_string"].codes, codes, strict=True)
        np.testing.assert_array_equal(mdict["cat_string"].categories, cats, strict=True)
        np.testing.assert_array_equal(mdict["cat_string"].ordered, ordered, strict=True)

    def test_categorical_3D(self, filename, version):
        """Test reading 3D categorical array data from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cat_3D"])
        assert "cat_3D" in mdict

        cats_list = sorted(["yes", "no", "maybe"])
        cats = np.array(cats_list, dtype=object)
        codes = np.tile(np.array([[2, 2], [1, 1], [0, 0]], dtype=np.int8), (2, 1, 1))
        ordered = False

        np.testing.assert_array_equal(mdict["cat_3D"].codes, codes, strict=True)
        np.testing.assert_array_equal(mdict["cat_3D"].categories, cats, strict=True)
        np.testing.assert_array_equal(mdict["cat_3D"].ordered, ordered, strict=True)


@pytest.mark.parametrize("filename, version", files)
class TestSaveMatlabCategorical:

    def test_categorical_scalar(self, filename, version):
        """Test writing categorical scalar to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cat_scalar"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["cat_scalar"])

            np.testing.assert_array_equal(
                mload["cat_scalar"].codes, mdict["cat_scalar"].codes, strict=True
            )
            np.testing.assert_array_equal(
                mload["cat_scalar"].categories,
                mdict["cat_scalar"].categories,
                strict=True,
            )
            np.testing.assert_array_equal(
                mload["cat_scalar"].ordered, mdict["cat_scalar"].ordered, strict=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_categorical_array(self, filename, version):
        """Test writing categorical array to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cat_array"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["cat_array"])

            np.testing.assert_array_equal(
                mload["cat_array"].codes, mdict["cat_array"].codes, strict=True
            )
            np.testing.assert_array_equal(
                mload["cat_array"].categories,
                mdict["cat_array"].categories,
                strict=True,
            )
            np.testing.assert_array_equal(
                mload["cat_array"].ordered, mdict["cat_array"].ordered, strict=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_categorical_empty(self, filename, version):
        """Test writing categorical empty to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cat_empty"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["cat_empty"])

            np.testing.assert_array_equal(
                mload["cat_empty"].codes, mdict["cat_empty"].codes, strict=True
            )
            np.testing.assert_array_equal(
                mload["cat_empty"].categories,
                mdict["cat_empty"].categories,
                strict=True,
            )
            np.testing.assert_array_equal(
                mload["cat_empty"].ordered, mdict["cat_empty"].ordered, strict=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_categorical_from_numeric(self, filename, version):
        """Test writing categorical numeric to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cat_from_numeric"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["cat_from_numeric"])

            np.testing.assert_array_equal(
                mload["cat_from_numeric"].codes,
                mdict["cat_from_numeric"].codes,
                strict=True,
            )
            np.testing.assert_array_equal(
                mload["cat_from_numeric"].categories,
                mdict["cat_from_numeric"].categories,
                strict=True,
            )
            np.testing.assert_array_equal(
                mload["cat_from_numeric"].ordered,
                mdict["cat_from_numeric"].ordered,
                strict=True,
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_categorical_unordered(self, filename, version):
        """Test writing categorical unordered to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cat_unordered"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["cat_unordered"])

            np.testing.assert_array_equal(
                mload["cat_unordered"].codes, mdict["cat_unordered"].codes, strict=True
            )
            np.testing.assert_array_equal(
                mload["cat_unordered"].categories,
                mdict["cat_unordered"].categories,
                strict=True,
            )
            np.testing.assert_array_equal(
                mload["cat_unordered"].ordered,
                mdict["cat_unordered"].ordered,
                strict=True,
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_categorical_ordered(self, filename, version):
        """Test writing categorical ordered to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cat_ordered"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["cat_ordered"])

            np.testing.assert_array_equal(
                mload["cat_ordered"].codes, mdict["cat_ordered"].codes, strict=True
            )
            np.testing.assert_array_equal(
                mload["cat_ordered"].categories,
                mdict["cat_ordered"].categories,
                strict=True,
            )
            np.testing.assert_array_equal(
                mload["cat_ordered"].ordered, mdict["cat_ordered"].ordered, strict=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_categorical_missing(self, filename, version):
        """Test writing categorical with missing values to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cat_missing"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["cat_missing"])

            np.testing.assert_array_equal(
                mload["cat_missing"].codes, mdict["cat_missing"].codes, strict=True
            )
            np.testing.assert_array_equal(
                mload["cat_missing"].categories,
                mdict["cat_missing"].categories,
                strict=True,
            )
            np.testing.assert_array_equal(
                mload["cat_missing"].ordered, mdict["cat_missing"].ordered, strict=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_categorical_mixed_case(self, filename, version):
        """Test writing categorical mixed case to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cat_mixed_case"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["cat_mixed_case"])

            np.testing.assert_array_equal(
                mload["cat_mixed_case"].codes,
                mdict["cat_mixed_case"].codes,
                strict=True,
            )
            np.testing.assert_array_equal(
                mload["cat_mixed_case"].categories,
                mdict["cat_mixed_case"].categories,
                strict=True,
            )
            np.testing.assert_array_equal(
                mload["cat_mixed_case"].ordered,
                mdict["cat_mixed_case"].ordered,
                strict=True,
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_categorical_matlab_string(self, filename, version):
        """Test writing categorical string to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cat_string"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["cat_string"])

            np.testing.assert_array_equal(
                mload["cat_string"].codes, mdict["cat_string"].codes, strict=True
            )
            np.testing.assert_array_equal(
                mload["cat_string"].categories,
                mdict["cat_string"].categories,
                strict=True,
            )
            np.testing.assert_array_equal(
                mload["cat_string"].ordered, mdict["cat_string"].ordered, strict=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_categorical_3D(self, filename, version):
        """Test writing categorical 3D to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cat_3D"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["cat_3D"])

            np.testing.assert_array_equal(
                mload["cat_3D"].codes, mdict["cat_3D"].codes, strict=True
            )
            np.testing.assert_array_equal(
                mload["cat_3D"].categories, mdict["cat_3D"].categories, strict=True
            )
            np.testing.assert_array_equal(
                mload["cat_3D"].ordered, mdict["cat_3D"].ordered, strict=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)



from pathlib import Path

import numpy as np
import pytest

from matio import whosmat

DATA_DIR = Path(__file__).parent / "data"


def get_files():
    """Get *.mat files from data directory (only base name)"""
    files = list(DATA_DIR.glob("*.mat"))
    bases_v7 = {f.stem[:-3] for f in files if f.stem.endswith("_v7")}
    bases = sorted(bases_v7)
    return bases


file_pairs = [
    (str(DATA_DIR / f"{base}_v7.mat"), str(DATA_DIR / f"{base}_v73.mat"))
    for base in get_files()
]


@pytest.mark.parametrize("file_v7, file_v73", file_pairs)
def test_whosmat(file_v7, file_v73):
    """Test whosmat function for both v7 and v7.3 files."""

    var_v7 = whosmat(file_v7)
    var_v73 = whosmat(file_v73)

    var_v7 = sorted(var_v7, key=lambda x: x[0])
    var_v73 = sorted(var_v73, key=lambda x: x[0])

    for var_v7, var_v73 in zip(var_v7, var_v73):
        name_v7, shape_v7, classname_v7 = var_v7
        name_v73, shape_v73, classname_v73 = var_v73

        if "sparse" in classname_v7 or "sparse" in classname_v73:
            continue  # Skip sparse matrices for now

        assert name_v7 == name_v73, f"Variable names differ: {name_v7} vs {name_v73}"
        assert (
            shape_v7 == shape_v73
        ), f"Shapes differ for {name_v7}: {shape_v7} vs {shape_v73}"
        assert (
            classname_v7 == classname_v73
        ), f"Class names differ for {name_v7}: {classname_v7} vs {classname_v73}"



"""Base class for MAT-file reading and writing"""

from scipy.sparse import coo_matrix, issparse

from matio.utils.matclass import MatReadError, MatWriteError
from matio.utils.matheaders import (
    MAT_5_VERSION,
    MAT_HDF_VERSION,
    MAT_VERSIONS,
    read_mat_header,
)
from matio.utils.matutils import sanitize_input_lists
from matio.v5 import loadmat5, savemat5, whosmat5
from matio.v7 import loadmat7, savemat7, whosmat7


def load_from_mat(
    file_path,
    mdict=None,
    variable_names=None,
    raw_data=False,
    add_table_attrs=False,
    spmatrix=True,
):
    """Loads variables from MAT-file"""

    subsystem_offset, ver, byte_order = read_mat_header(file_path)
    variable_names = sanitize_input_lists(variable_names, "variable_names")

    if ver == MAT_5_VERSION:
        matfile_dict = loadmat5(
            file_path,
            subsystem_offset,
            byte_order,
            variable_names,
            raw_data,
            add_table_attrs,
        )
    elif ver == MAT_HDF_VERSION:
        matfile_dict = loadmat7(
            file_path, byte_order, variable_names, raw_data, add_table_attrs
        )

    if len(matfile_dict["__globals__"]) == 0:
        del matfile_dict["__globals__"]

    if spmatrix:
        for name, var in list(matfile_dict.items()):
            if issparse(var):
                matfile_dict[name] = coo_matrix(var)

    # Update mdict if present
    if mdict is not None:
        mdict.update(matfile_dict)
    else:
        mdict = matfile_dict

    return mdict


def whosmat(file_path):
    """Lists variables in MAT-file"""

    _, ver, byte_order = read_mat_header(file_path)

    if ver == MAT_5_VERSION:
        vars = whosmat5(file_path, byte_order)
    elif ver == MAT_HDF_VERSION:
        vars = whosmat7(file_path)

    return vars


def save_to_mat(
    file_path,
    mdict,
    version="v7.3",
    global_vars=None,
    saveobj_classes=None,
    oned_as="col",
    do_compression=True,
):
    """Saves variables to MAT-file"""

    global_vars = sanitize_input_lists(global_vars, "global_vars")
    saveobj_classes = sanitize_input_lists(saveobj_classes, "saveobj_classes")

    ver_int = MAT_VERSIONS.get(version)
    if ver_int is None:
        raise MatWriteError(
            f"Unknown MAT-file version '{version}' specified. Supported versions are: {list(MAT_VERSIONS.keys())}"
        )

    if ver_int == MAT_5_VERSION:
        savemat5(
            file_path, mdict, global_vars, saveobj_classes, oned_as, do_compression
        )
    elif ver_int == MAT_HDF_VERSION:
        savemat7(file_path, mdict, global_vars, saveobj_classes, oned_as)



"""MAT-file 5 reading and writing"""

# Copyright (c) 2001-2002 Enthought, Inc. 2003, SciPy Developers.
# All rights reserved.
#
# Modified by foreverallama (c) 2025
# https://github.com/foreverallama/matio
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import math
import sys
import warnings
import zlib
from io import BytesIO

import numpy as np
import scipy.sparse

from matio.subsystem import MatSubsystem
from matio.utils.matclass import (
    MCOS_SUBSYSTEM_CLASS,
    EmptyMatStruct,
    MatlabCanonicalEmpty,
    MatlabEnumerationArray,
    MatlabFunction,
    MatlabObject,
    MatlabOpaque,
    MatlabOpaqueArray,
    MatWriteError,
    MatWriteWarning,
)
from matio.utils.matheaders import (
    MAT5_MAX_ARR_BYTES,
    MAT5_MAX_STRUCT_FIELDNAME_LEN,
    MAT_5_VERSION,
    write_file_header,
    write_subsystem_offset,
    write_version,
)
from matio.utils.matutils import (
    _get_string_arr_dtype,
    mat_numeric,
    matdims,
    strings_to_chars,
    to_writeable,
)

# Constants and helper objects
from matio.v5.matio5_params import (
    MDTYPES,
    NP_TO_MTYPES,
    NP_TO_MXTYPES,
    miTypes,
    mxTypes,
)

SYS_BYTE_ORDER = "<" if sys.byteorder == "little" else ">"

# Native byte ordered dtypes for convenience for writers
NDT_FILE_HDR = MDTYPES[SYS_BYTE_ORDER]["dtypes"]["file_header"]
NDT_TAG_FULL = MDTYPES[SYS_BYTE_ORDER]["dtypes"]["tag_full"]
NDT_TAG_SMALL = MDTYPES[SYS_BYTE_ORDER]["dtypes"]["tag_smalldata"]
NDT_ARRAY_FLAGS = MDTYPES[SYS_BYTE_ORDER]["dtypes"]["array_flags"]


def savemat5(file_path, mdict, global_vars, saveobj_classes, oned_as, do_compression):
    """Write data to MAT-5file."""

    with open(file_path, "wb") as f:
        write_file_header(f, version=MAT_5_VERSION)
        MW = MatFile5Writer(f, oned_as=oned_as)
        MW.subsystem = MatSubsystem(
            byte_order=SYS_BYTE_ORDER, oned_as=oned_as, saveobj_classes=saveobj_classes
        )
        MW.subsystem.init_save()

        MW.put_variables(mdict, global_vars, do_compression)
        subsystem = MW.subsystem.set_subsystem()
        if subsystem is None:
            return

        subsystem_offset = f.tell()
        subsys_stream = write_subsystem(subsystem, oned_as)

        temp_dict = {"__subsystem__": subsys_stream}
        MW.put_variables(temp_dict, global_vars=[], do_compression=do_compression)

        write_subsystem_offset(f, subsystem_offset)

    return


def write_subsystem(subsystem, oned_as):
    """Write subsystem data to subsys stream"""

    subsys_stream = BytesIO()
    subsys_stream.seek(0)
    write_version(subsys_stream, MAT_5_VERSION)
    subsys_stream.write(b"\x00" * 4)  # Padding

    MW = MatFile5Writer(subsys_stream, oned_as=oned_as)
    temp_dict = {"__subsystem__": subsystem}

    MW.put_variables(temp_dict, global_vars=[], do_compression=False)
    subsys_view = np.frombuffer(subsys_stream.getbuffer(), dtype=np.uint8)
    return subsys_view


class VarWriter5:
    """Generic matlab matrix writing class"""

    mat_tag = np.zeros((), NDT_TAG_FULL)
    mat_tag["mdtype"] = miTypes.miMATRIX

    def __init__(self, file_writer):
        self.file_stream = file_writer.file_stream
        self.oned_as = file_writer.oned_as
        self.subsystem = file_writer.subsystem

        # These are used for top level writes, and unset after
        self._var_name = None
        self._var_is_global = False

    def write_bytes(self, arr):
        self.file_stream.write(arr.tobytes(order="F"))

    def write_string(self, s):
        self.file_stream.write(s)

    def write_element(self, arr, mdtype=None):
        """write tag and data"""
        if mdtype is None:
            mdtype = NP_TO_MTYPES[arr.dtype.str[1:]]
        # Array needs to be in native byte order
        if arr.dtype.byteorder == (SYS_BYTE_ORDER == ">"):
            arr = arr.byteswap().view(arr.dtype.newbyteorder())
        byte_count = arr.size * arr.itemsize
        if byte_count <= 4:
            self.write_smalldata_element(arr, mdtype, byte_count)
        else:
            self.write_regular_element(arr, mdtype, byte_count)

    def write_smalldata_element(self, arr, mdtype, byte_count):
        # write tag with embedded data
        tag = np.zeros((), NDT_TAG_SMALL)
        tag["byte_count_mdtype"] = (byte_count << 16) + mdtype
        # if arr.tobytes is < 4, the element will be zero-padded as needed.
        tag["data"] = arr.tobytes(order="F")
        self.write_bytes(tag)

    def write_regular_element(self, arr, mdtype, byte_count):
        # write tag, data
        tag = np.zeros((), NDT_TAG_FULL)
        tag["mdtype"] = mdtype
        tag["byte_count"] = byte_count
        self.write_bytes(tag)
        self.write_bytes(arr)
        # pad to next 64-bit boundary
        bc_mod_8 = byte_count % 8
        if bc_mod_8:
            self.file_stream.write(b"\x00" * (8 - bc_mod_8))

    def write_header(self, shape, mclass, is_complex=False, is_logical=False, nzmax=0):
        """Write header for given data options
        shape : sequence
           array shape
        mclass      - mat5 matrix class
        is_complex  - True if matrix is complex
        is_logical  - True if matrix is logical
        nzmax        - max non zero elements for sparse arrays

        We get the name and the global flag from the object, and reset
        them to defaults after we've used them
        """
        # get name and is_global from one-shot object store
        name = self._var_name
        is_global = self._var_is_global
        # initialize the top-level matrix tag, store position
        self._mat_tag_pos = self.file_stream.tell()
        self.write_bytes(self.mat_tag)
        # write array flags (complex, global, logical, class, nzmax)
        af = np.zeros((), NDT_ARRAY_FLAGS)
        af["data_type"] = miTypes.miUINT32
        af["byte_count"] = 8
        flags = is_complex << 3 | is_global << 2 | is_logical << 1
        af["flags_class"] = mclass | flags << 8
        af["nzmax"] = nzmax
        self.write_bytes(af)
        # shape
        if shape is not None:
            self.write_element(np.array(shape, dtype="i4"))
        # write name
        name = np.asarray(name)
        if name == "" or name == b"":  # empty string zero-terminated
            self.write_smalldata_element(name, miTypes.miINT8, 0)
        else:
            self.write_element(name, miTypes.miINT8)
        # reset the one-shot store to defaults
        self._var_name = ""
        self._var_is_global = False

    def update_matrix_tag(self, start_pos):
        curr_pos = self.file_stream.tell()
        self.file_stream.seek(start_pos)
        byte_count = curr_pos - start_pos - 8
        if byte_count >= MAT5_MAX_ARR_BYTES:
            raise MatWriteError(
                "Matrix too large to save with MAT-5 file format. Use v7.3 format instead."
            )
        self.mat_tag["byte_count"] = byte_count
        self.write_bytes(self.mat_tag)
        self.file_stream.seek(curr_pos)

    def write_top(self, arr, name, is_global):
        """Write top-level variable"""
        # these are set before the top-level header write, and unset at
        # the end of the same write, because they do not apply for lower levels
        self._var_is_global = is_global
        self._var_name = name
        self.write(arr)

    def write(self, arr):
        """Write data element to MAT-file"""

        mat_tag_pos = self.file_stream.tell()

        narr = to_writeable(arr, self.oned_as)

        if scipy.sparse.issparse(narr):
            self.write_sparse(narr)
        elif isinstance(narr, MatlabCanonicalEmpty):
            self.write_canonical_empty(narr)
        elif isinstance(narr, MatlabObject):
            self.write_object(narr)
        elif isinstance(narr, MatlabFunction):
            self.write_function_handle(narr)
        elif isinstance(narr, EmptyMatStruct):
            self.write_empty_struct(narr)
        elif isinstance(
            narr, (MatlabOpaque, MatlabOpaqueArray, MatlabEnumerationArray)
        ):
            self.write_opaque(narr)
        elif narr.dtype.fields:  # struct array
            self.write_struct(narr)
        elif narr.dtype.hasobject:  # cell array
            self.write_cells(narr)
        elif narr.dtype.kind in ("U", "S"):
            self.write_char(narr)
        elif narr.dtype.kind in ("i", "u", "f", "c", "b"):
            self.write_numeric(narr)

        self.update_matrix_tag(mat_tag_pos)

    def write_numeric(self, arr):
        is_complex = arr.dtype.kind == "c"
        is_logical = arr.dtype.kind == "b"

        arr, _, _ = mat_numeric(arr, version=MAT_5_VERSION)
        mclass = NP_TO_MXTYPES[arr.dtype.str[1:]]

        self.write_header(
            matdims(arr, self.oned_as),
            mclass,
            is_complex=is_complex,
            is_logical=is_logical,
        )

        if is_complex:
            self.write_element(arr.real)
            self.write_element(arr.imag)
        else:
            self.write_element(arr)

    def write_char(self, arr, codec="ascii"):
        """Write char array (from numpy string array)"""

        if arr.size == 0 or np.all(arr == ""):
            # Empty String Array
            # MATLAB only recognizes 0-D char array
            shape = (0,) * np.max([arr.ndim, 2])
            self.write_header(shape, mxTypes.mxCHAR_CLASS)
            self.write_smalldata_element(arr, miTypes.miUTF8, 0)
            return

        # Convert string to char array
        arr = strings_to_chars(arr)
        # We have to write the shape directly, because we are going
        # recode the characters, and the resulting stream of chars
        # may have a different length
        shape = arr.shape
        self.write_header(shape, mxTypes.mxCHAR_CLASS)
        if arr.dtype.kind == "U" and arr.size:
            # Make one long string from all the characters. We need to
            # transpose here, because we're flattening the array, before
            # we write the bytes. The bytes have to be written in
            # Fortran order.
            n_chars = math.prod(shape)
            st_arr = np.ndarray(
                shape=(), dtype=_get_string_arr_dtype(arr, n_chars), buffer=arr.T.copy()
            )  # Fortran order
            # Recode with codec to give byte string
            st = st_arr.item().encode(codec)
            # Reconstruct as 1-D byte array
            arr = np.ndarray(shape=(len(st),), dtype="S1", buffer=st)
        self.write_element(arr, mdtype=miTypes.miUTF8)

    def write_sparse(self, arr):
        """Sparse matrices are 2D"""
        A = arr.tocsc()  # convert to sparse CSC format
        A.sort_indices()  # MATLAB expects sorted row indices
        is_complex = A.dtype.kind == "c"
        is_logical = A.dtype.kind == "b"
        nz = A.nnz
        self.write_header(
            matdims(arr, self.oned_as),
            mxTypes.mxSPARSE_CLASS,
            is_complex=is_complex,
            is_logical=is_logical,
            # matlab won't load file with 0 nzmax
            nzmax=1 if nz == 0 else nz,
        )
        self.write_element(A.indices.astype("i4"))
        self.write_element(A.indptr.astype("i4"))
        self.write_element(A.data.real)
        if is_complex:
            self.write_element(A.data.imag)

    def write_cells(self, arr):
        """Write cell array"""
        self.write_header(matdims(arr, self.oned_as), mxTypes.mxCELL_CLASS)
        # loop over data, column major
        A = np.atleast_2d(arr).flatten("F")
        for el in A:
            self.write(el)

    def write_empty_struct(self, arr):
        """Write empty struct array"""
        self.write_header(arr.shape, mxTypes.mxSTRUCT_CLASS)
        # max field name length set to 1 in an example matlab struct
        self.write_element(np.array(1, dtype=np.int32))
        # Field names element is empty
        self.write_element(np.array([], dtype=np.int8))

    def write_struct(self, arr):
        """Write struct array"""
        self.write_header(matdims(arr, self.oned_as), mxTypes.mxSTRUCT_CLASS)
        self._write_items(arr)

    def _write_items(self, arr):
        """Write struct fields"""
        # write fieldnames
        fieldnames = [f[0] for f in arr.dtype.descr]
        length = max([len(fieldname) for fieldname in fieldnames]) + 1
        if length > MAT5_MAX_STRUCT_FIELDNAME_LEN:
            raise ValueError(
                f"Field names are restricted to {MAT5_MAX_STRUCT_FIELDNAME_LEN - 1} characters"
            )
        self.write_element(np.array([length], dtype="i4"))
        self.write_element(
            np.array(fieldnames, dtype=f"S{length}"), mdtype=miTypes.miINT8
        )
        A = np.atleast_2d(arr).flatten("F")
        for el in A:
            for f in fieldnames:
                self.write(el[f])

    def write_object(self, arr):
        """Write MatlabObject"""
        self.write_header(matdims(arr, self.oned_as), mxTypes.mxOBJECT_CLASS)
        self.write_element(np.array(arr.classname, dtype="S"), mdtype=miTypes.miINT8)
        self._write_items(arr)

    def write_function_handle(self, arr):
        """Write MatlabFunction"""
        self.write_header(matdims(arr, self.oned_as), mxTypes.mxFUNCTION_CLASS)
        self.write(arr.view(np.ndarray))

    def write_opaque(self, arr):
        """Array Flags, Var Name, Type System, Class Name, Metadata"""
        self.write_header(None, mxTypes.mxOPAQUE_CLASS)
        self.write_element(np.array(arr.type_system, dtype="S"), mdtype=miTypes.miINT8)
        self.write_element(np.array(arr.classname, dtype="S"), mdtype=miTypes.miINT8)
        if arr.classname == MCOS_SUBSYSTEM_CLASS:
            self.write(arr.properties)
        else:
            if isinstance(arr, MatlabEnumerationArray):
                objmetadata = self.subsystem.set_enumeration_metadata(arr)
            else:
                objmetadata = self.subsystem.set_object_metadata(arr)
            self.write(objmetadata)

    def write_canonical_empty(self, arr):
        """Write canonical empty array"""
        self._mat_tag_pos = self.file_stream.tell()
        self.write_bytes(self.mat_tag)  # 0 byte miMATRIX Tag


class MatFile5Writer:
    """Class for writing mat5 files"""

    def __init__(self, file_stream, oned_as):
        """Initialize writer for MAT-5 format"""
        self.file_stream = file_stream
        self.oned_as = oned_as
        self._matrix_writer = None
        self.subsystem = None

    def put_variables(self, mdict, global_vars, do_compression):
        """Write variables to MATLAB file."""

        self._matrix_writer = VarWriter5(self)

        for name, var in mdict.items():
            if name == "__subsystem__":
                name = ""
            elif name[0] in "_0123456789":
                msg = f"Variable names cannot start with '{name[0]}'. Skipping '{name}'"
                warnings.warn(msg, MatWriteWarning, stacklevel=2)
                continue
            is_global = name in global_vars
            if do_compression:
                stream = BytesIO()
                self._matrix_writer.file_stream = stream
                self._matrix_writer.write_top(var, name.encode("latin1"), is_global)
                out_str = zlib.compress(stream.getvalue())
                tag = np.empty((), NDT_TAG_FULL)
                tag["mdtype"] = miTypes.miCOMPRESSED
                tag["byte_count"] = len(out_str)
                self.file_stream.write(tag.tobytes())
                self.file_stream.write(out_str)
            else:
                self._matrix_writer.write_top(var, name.encode("latin1"), is_global)



import os
import tempfile

import numpy as np
import pytest

from matio import load_from_mat, save_to_mat
from matio.utils.matclass import MatlabEnumerationArray, MatlabOpaque

files = [("test_enum_v7.mat", "v7"), ("test_enum_v73.mat", "v7.3")]
namespace = "TestClasses"


@pytest.mark.parametrize("filename, version", files)
class TestLoadMatlabEnum:

    def test_enum_scalar(self, filename, version):
        """Test reading enum scalar from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["enum_scalar"])
        assert "enum_scalar" in mdict
        assert isinstance(mdict["enum_scalar"], MatlabEnumerationArray)
        assert mdict["enum_scalar"].classname == f"{namespace}.EnumClass"
        assert mdict["enum_scalar"].type_system == "MCOS"
        assert mdict["enum_scalar"].shape == (1, 1)
        assert mdict["enum_scalar"].dtype == object

        assert mdict["enum_scalar"][0, 0].name == "enum1"
        val_dict = {
            "val": np.array([[1]], dtype=np.float64),
        }
        for key, val in val_dict.items():
            assert key in mdict["enum_scalar"][0, 0].value
            np.testing.assert_array_equal(mdict["enum_scalar"][0, 0].value[key], val)

    def test_enum_uint32(self, filename, version):
        """Test reading enum array from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["enum_uint32"])
        assert "enum_uint32" in mdict
        assert isinstance(mdict["enum_uint32"], MatlabEnumerationArray)
        assert mdict["enum_uint32"].classname == f"{namespace}.EnumClassWithBase"
        assert mdict["enum_uint32"].type_system == "MCOS"
        assert mdict["enum_uint32"].shape == (1, 1)
        assert mdict["enum_uint32"].dtype == object

        assert mdict["enum_uint32"][0, 0].name == "enum1"
        val_dict = {
            "uint32.Data": np.array([[1]], dtype=np.uint32),
        }
        for key, val in val_dict.items():
            assert key in mdict["enum_uint32"][0, 0].value
            np.testing.assert_array_equal(mdict["enum_uint32"][0, 0].value[key], val)

    def test_enum_array(self, filename, version):
        """Test reading datetime array data from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["enum_array"])
        assert "enum_array" in mdict
        assert isinstance(mdict["enum_array"], MatlabEnumerationArray)
        assert mdict["enum_array"].classname == f"{namespace}.EnumClass"
        assert mdict["enum_array"].type_system == "MCOS"
        assert mdict["enum_array"].shape == (2, 3)
        assert mdict["enum_array"].dtype == object

        expected_names = np.array(
            [["enum1", "enum3", "enum5"], ["enum2", "enum4", "enum6"]], dtype=np.str_
        ).reshape((2, 3), order="F")

        expected_vals = np.array([[1, 3, 5], [2, 4, 6]], dtype=np.float64).reshape(
            (2, 3), order="F"
        )

        for idx in np.ndindex(mdict["enum_array"].shape):
            enum_obj = mdict["enum_array"][idx]
            assert enum_obj.name == expected_names[idx]
            assert "val" in enum_obj.value
            np.testing.assert_array_equal(
                enum_obj.value["val"],
                np.array([[expected_vals[idx]]], dtype=np.float64),
            )

    def test_load_enum_nested(self, filename, version):
        """Test reading nested enum data from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["enum_nested"])
        assert "enum_nested" in mdict
        assert isinstance(mdict["enum_nested"], MatlabOpaque)
        assert mdict["enum_nested"].classname == f"{namespace}.BasicClass"
        assert mdict["enum_nested"].type_system == "MCOS"

        assert isinstance(mdict["enum_nested"].properties["a"], MatlabEnumerationArray)
        assert (
            mdict["enum_nested"].properties["a"].classname == f"{namespace}.EnumClass"
        )
        assert mdict["enum_nested"].properties["a"].type_system == "MCOS"
        assert mdict["enum_nested"].properties["a"].shape == (1, 1)
        assert mdict["enum_nested"].properties["a"].dtype == object
        assert mdict["enum_nested"].properties["a"][0, 0].name == "enum1"

        assert isinstance(mdict["enum_nested"].properties["b"], np.ndarray)
        assert isinstance(
            mdict["enum_nested"].properties["b"][0, 0], MatlabEnumerationArray
        )
        assert (
            mdict["enum_nested"].properties["b"][0, 0].classname
            == f"{namespace}.EnumClass"
        )
        assert mdict["enum_nested"].properties["b"][0, 0].type_system == "MCOS"
        assert mdict["enum_nested"].properties["b"][0, 0].shape == (1, 1)
        assert mdict["enum_nested"].properties["b"][0, 0].dtype == object
        assert mdict["enum_nested"].properties["b"][0, 0][0, 0].name == "enum2"

        assert isinstance(mdict["enum_nested"].properties["c"], np.ndarray)
        assert isinstance(
            mdict["enum_nested"].properties["c"][0, 0]["InnerProp"],
            MatlabEnumerationArray,
        )
        assert (
            mdict["enum_nested"].properties["c"][0, 0]["InnerProp"].classname
            == f"{namespace}.EnumClass"
        )
        assert (
            mdict["enum_nested"].properties["c"][0, 0]["InnerProp"].type_system
            == "MCOS"
        )
        assert mdict["enum_nested"].properties["c"][0, 0]["InnerProp"].shape == (1, 1)
        assert mdict["enum_nested"].properties["c"][0, 0]["InnerProp"].dtype == object
        assert (
            mdict["enum_nested"].properties["c"][0, 0]["InnerProp"][0, 0].name
            == "enum3"
        )


@pytest.mark.parametrize("filename, version", files)
class TestSaveMatlabEnum:

    def test_enum_scalar(self, filename, version):
        """Test writing enum scalar to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["enum_scalar"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["enum_scalar"])
            assert "enum_scalar" in mload
            assert isinstance(mload["enum_scalar"], MatlabEnumerationArray)
            assert mload["enum_scalar"].classname == mdict["enum_scalar"].classname
            assert mload["enum_scalar"].type_system == mdict["enum_scalar"].type_system
            assert mload["enum_scalar"].shape == mdict["enum_scalar"].shape
            assert mload["enum_scalar"].dtype == mdict["enum_scalar"].dtype

            assert mload["enum_scalar"][0, 0].name == mdict["enum_scalar"][0, 0].name
            for key in mdict["enum_scalar"][0, 0].value:
                assert key in mload["enum_scalar"][0, 0].value
                np.testing.assert_array_equal(
                    mload["enum_scalar"][0, 0].value[key],
                    mdict["enum_scalar"][0, 0].value[key],
                )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_enum_uint32(self, filename, version):
        """Test writing enum array to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["enum_uint32"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["enum_uint32"])
            assert "enum_uint32" in mload
            assert isinstance(mload["enum_uint32"], MatlabEnumerationArray)
            assert mload["enum_uint32"].classname == mdict["enum_uint32"].classname
            assert mload["enum_uint32"].type_system == mdict["enum_uint32"].type_system
            assert mload["enum_uint32"].shape == mdict["enum_uint32"].shape
            assert mload["enum_uint32"].dtype == mdict["enum_uint32"].dtype

            assert mload["enum_uint32"][0, 0].name == mdict["enum_uint32"][0, 0].name
            for key in mdict["enum_uint32"][0, 0].value:
                assert key in mload["enum_uint32"][0, 0].value
                np.testing.assert_array_equal(
                    mload["enum_uint32"][0, 0].value[key],
                    mdict["enum_uint32"][0, 0].value[key],
                )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_enum_array(self, filename, version):
        """Test writing enum array data to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["enum_array"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["enum_array"])
            assert "enum_array" in mload
            assert isinstance(mload["enum_array"], MatlabEnumerationArray)
            assert mload["enum_array"].classname == mdict["enum_array"].classname
            assert mload["enum_array"].type_system == mdict["enum_array"].type_system
            assert mload["enum_array"].shape == mdict["enum_array"].shape
            assert mload["enum_array"].dtype == mdict["enum_array"].dtype

            for idx in np.ndindex(mdict["enum_array"].shape):
                enum_obj_orig = mdict["enum_array"][idx]
                enum_obj_load = mload["enum_array"][idx]
                assert enum_obj_load.name == enum_obj_orig.name
                for key in enum_obj_orig.value:
                    assert key in enum_obj_load.value
                    np.testing.assert_array_equal(
                        enum_obj_load.value[key],
                        enum_obj_orig.value[key],
                    )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_save_enum_nested(self, filename, version):
        """Test writing nested enum data to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["enum_nested"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["enum_nested"])
            assert "enum_nested" in mload
            assert isinstance(mload["enum_nested"], MatlabOpaque)
            assert mload["enum_nested"].classname == mdict["enum_nested"].classname
            assert mload["enum_nested"].type_system == mdict["enum_nested"].type_system

            assert isinstance(
                mload["enum_nested"].properties["a"], MatlabEnumerationArray
            )
            assert (
                mload["enum_nested"].properties["a"].classname
                == mdict["enum_nested"].properties["a"].classname
            )
            assert (
                mload["enum_nested"].properties["a"].type_system
                == mdict["enum_nested"].properties["a"].type_system
            )
            assert (
                mload["enum_nested"].properties["a"].shape
                == mdict["enum_nested"].properties["a"].shape
            )
            assert (
                mload["enum_nested"].properties["a"].dtype
                == mdict["enum_nested"].properties["a"].dtype
            )
            assert (
                mload["enum_nested"].properties["a"][0, 0].name
                == mdict["enum_nested"].properties["a"][0, 0].name
            )

            assert isinstance(mload["enum_nested"].properties["b"], np.ndarray)
            assert isinstance(
                mload["enum_nested"].properties["b"][0, 0], MatlabEnumerationArray
            )
            assert (
                mload["enum_nested"].properties["b"][0, 0].classname
                == mdict["enum_nested"].properties["b"][0, 0].classname
            )
            assert (
                mload["enum_nested"].properties["b"][0, 0].type_system
                == mdict["enum_nested"].properties["b"][0, 0].type_system
            )
            assert (
                mload["enum_nested"].properties["b"][0, 0].shape
                == mdict["enum_nested"].properties["b"][0, 0].shape
            )
            assert (
                mload["enum_nested"].properties["b"][0, 0].dtype
                == mdict["enum_nested"].properties["b"][0, 0].dtype
            )
            assert (
                mload["enum_nested"].properties["b"][0, 0][0, 0].name
                == mdict["enum_nested"].properties["b"][0, 0][0, 0].name
            )
            assert isinstance(mload["enum_nested"].properties["c"], np.ndarray)
            assert isinstance(
                mload["enum_nested"].properties["c"][0, 0]["InnerProp"],
                MatlabEnumerationArray,
            )
            assert (
                mload["enum_nested"].properties["c"][0, 0]["InnerProp"].classname
                == mdict["enum_nested"].properties["c"][0, 0]["InnerProp"].classname
            )
            assert (
                mload["enum_nested"].properties["c"][0, 0]["InnerProp"].type_system
                == mdict["enum_nested"].properties["c"][0, 0]["InnerProp"].type_system
            )
            assert (
                mload["enum_nested"].properties["c"][0, 0]["InnerProp"].shape
                == mdict["enum_nested"].properties["c"][0, 0]["InnerProp"].shape
            )
            assert (
                mload["enum_nested"].properties["c"][0, 0]["InnerProp"].dtype
                == mdict["enum_nested"].properties["c"][0, 0]["InnerProp"].dtype
            )
            assert (
                mload["enum_nested"].properties["c"][0, 0]["InnerProp"][0, 0].name
                == mdict["enum_nested"].properties["c"][0, 0]["InnerProp"][0, 0].name
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)



import os
import tempfile

import numpy as np
import pytest

from matio import load_from_mat, save_to_mat
from matio.utils.matclass import MatWriteWarning

files = [("test_basic_v7.mat", "v7"), ("test_basic_v73.mat", "v7.3")]


@pytest.mark.parametrize("filename, version", files)
class TestLoadBasicDataTypes:

    def test_numeric_int8(self, filename, version):
        """Test reading int8 data from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["int8_scalar", "int8_array"])
        assert "int8_scalar" in mdict
        assert "int8_array" in mdict

        int8_scalar = np.array([[42]], dtype=np.int8)
        int8_array = np.array([[1, 2, 3, 4, 5, 6]], dtype=np.int8).reshape(2, 3)

        np.testing.assert_array_equal(mdict["int8_scalar"], int8_scalar, strict=True)
        np.testing.assert_array_equal(mdict["int8_array"], int8_array, strict=True)

    def test_numeric_uint8(self, filename, version):
        """Test reading uint8 data from MAT-file"""

        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["uint8_scalar", "uint8_array"])
        assert "uint8_scalar" in mdict
        assert "uint8_array" in mdict

        uint8_scalar = np.array([[42]], dtype=np.uint8)
        uint8_array = np.array([[1, 2, 3, 4, 5, 6]], dtype=np.uint8).reshape(2, 3)

        np.testing.assert_array_equal(mdict["uint8_scalar"], uint8_scalar, strict=True)
        np.testing.assert_array_equal(mdict["uint8_array"], uint8_array, strict=True)

    def test_numeric_int16(self, filename, version):
        """Test reading int16 data from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["int16_scalar", "int16_array"])
        assert "int16_scalar" in mdict
        assert "int16_array" in mdict

        int16_scalar = np.array([[42]], dtype=np.int16)
        int16_array = np.array([[1, 2, 3, 4, 5, 6]], dtype=np.int16).reshape(2, 3)

        np.testing.assert_array_equal(mdict["int16_scalar"], int16_scalar, strict=True)
        np.testing.assert_array_equal(mdict["int16_array"], int16_array, strict=True)

    def test_numeric_uint16(self, filename, version):
        """Test reading uint16 data from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(
            file_path, variable_names=["uint16_scalar", "uint16_array"]
        )
        assert "uint16_scalar" in mdict
        assert "uint16_array" in mdict

        uint16_scalar = np.array([[42]], dtype=np.uint16)
        uint16_array = np.array([[1, 2, 3, 4, 5, 6]], dtype=np.uint16).reshape(2, 3)

        np.testing.assert_array_equal(
            mdict["uint16_scalar"], uint16_scalar, strict=True
        )
        np.testing.assert_array_equal(mdict["uint16_array"], uint16_array, strict=True)

    def test_numeric_int32(self, filename, version):
        """Test reading int32 data from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["int32_scalar", "int32_array"])
        assert "int32_scalar" in mdict
        assert "int32_array" in mdict

        int32_scalar = np.array([[42]], dtype=np.int32)
        int32_array = np.array([[1, 2, 3, 4, 5, 6]], dtype=np.int32).reshape(2, 3)

        np.testing.assert_array_equal(mdict["int32_scalar"], int32_scalar, strict=True)
        np.testing.assert_array_equal(mdict["int32_array"], int32_array, strict=True)

    def test_numeric_uint32(self, filename, version):
        """Test reading uint32 data from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(
            file_path, variable_names=["uint32_scalar", "uint32_array"]
        )
        assert "uint32_scalar" in mdict
        assert "uint32_array" in mdict

        uint32_scalar = np.array([[42]], dtype=np.uint32)
        uint32_array = np.array([[1, 2, 3, 4, 5, 6]], dtype=np.uint32).reshape(2, 3)

        np.testing.assert_array_equal(
            mdict["uint32_scalar"], uint32_scalar, strict=True
        )
        np.testing.assert_array_equal(mdict["uint32_array"], uint32_array, strict=True)

    def test_numeric_int64(self, filename, version):
        """Test reading int64 data from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["int64_scalar", "int64_array"])
        assert "int64_scalar" in mdict
        assert "int64_array" in mdict

        int64_scalar = np.array([[42]], dtype=np.int64)
        int64_array = np.array([[1, 2, 3, 4, 5, 6]], dtype=np.int64).reshape(2, 3)

        np.testing.assert_array_equal(mdict["int64_scalar"], int64_scalar, strict=True)
        np.testing.assert_array_equal(mdict["int64_array"], int64_array, strict=True)

    def test_numeric_uint64(self, filename, version):
        """Test reading uint64 data from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(
            file_path, variable_names=["uint64_scalar", "uint64_array"]
        )
        assert "uint64_scalar" in mdict
        assert "uint64_array" in mdict

        uint64_scalar = np.array([[42]], dtype=np.uint64)
        uint64_array = np.array([[1, 2, 3, 4, 5, 6]], dtype=np.uint64).reshape(2, 3)

        np.testing.assert_array_equal(
            mdict["uint64_scalar"], uint64_scalar, strict=True
        )
        np.testing.assert_array_equal(mdict["uint64_array"], uint64_array, strict=True)

    def test_numeric_single(self, filename, version):
        """Test reading single data from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(
            file_path, variable_names=["single_scalar", "single_array"]
        )
        assert "single_scalar" in mdict
        assert "single_array" in mdict

        single_scalar = np.array([[3.14]], dtype=np.float32)
        single_array = np.array(
            [[1.1, 2.2, 3.3, 4.4, 5.5, 6.6]], dtype=np.float32
        ).reshape(2, 3)

        np.testing.assert_array_equal(
            mdict["single_scalar"], single_scalar, strict=True
        )
        np.testing.assert_array_equal(mdict["single_array"], single_array, strict=True)

    def test_numeric_double(self, filename, version):
        """Test reading double data from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(
            file_path, variable_names=["double_scalar", "double_array"]
        )
        assert "double_scalar" in mdict
        assert "double_array" in mdict

        double_scalar = np.array([[3.14]], dtype=np.float64)
        double_array = np.array(
            [[1.1, 2.2, 3.3, 4.4, 5.5, 6.6]], dtype=np.float64
        ).reshape(2, 3)

        np.testing.assert_array_equal(
            mdict["double_scalar"], double_scalar, strict=True
        )
        np.testing.assert_array_equal(mdict["double_array"], double_array, strict=True)

    def test_logical(self, filename, version):
        """Test reading logical data from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(
            file_path, variable_names=["logical_scalar", "logical_array"]
        )
        assert "logical_scalar" in mdict
        assert "logical_array" in mdict

        logical_scalar = np.array([[1]], dtype=bool)
        logical_array = np.array([[1, 0, 1]], dtype=bool).reshape(1, 3)

        np.testing.assert_array_equal(
            mdict["logical_scalar"], logical_scalar, strict=True
        )
        np.testing.assert_array_equal(
            mdict["logical_array"], logical_array, strict=True
        )

    def test_complex(self, filename, version):
        """Test reading complex data from MAT-file"""

        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(
            file_path, variable_names=["complex_scalar", "complex_array"]
        )
        assert "complex_scalar" in mdict
        assert "complex_array" in mdict

        complex_scalar = np.array([[1.0 + 2.0j]], dtype=np.complex128).reshape(1, 1)
        complex_array = np.array(
            [[1.0 + 2.0j, 2.0 + 4.0j, 4.0 + 8.0j]], dtype=np.complex128
        ).reshape(3, 1)

        np.testing.assert_array_equal(
            mdict["complex_scalar"], complex_scalar, strict=True
        )
        np.testing.assert_array_equal(
            mdict["complex_array"], complex_array, strict=True
        )

    def test_char(self, filename, version):
        """Test reading char data from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["char_array", "char_scalar"])
        assert "char_array" in mdict
        assert "char_scalar" in mdict

        char_scalar = np.array(["Hello"], dtype=np.str_).reshape(
            1,
        )
        char_array = np.array(["ab", "cd", "ef"], dtype=np.str_).reshape(
            3,
        )

        np.testing.assert_array_equal(mdict["char_scalar"], char_scalar, strict=True)
        np.testing.assert_array_equal(mdict["char_array"], char_array, strict=True)

    def test_empty(self, filename, version):
        """Test reading empty arrays from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(
            file_path, variable_names=["numeric_empty", "char_empty", "logical_empty"]
        )
        assert "numeric_empty" in mdict
        assert "char_empty" in mdict
        assert "logical_empty" in mdict

        numeric_empty = np.empty((0, 0))
        char_empty = np.empty((0,), dtype=np.str_)
        logical_empty = np.empty((0, 0), dtype=np.bool_)

        np.testing.assert_array_equal(
            mdict["numeric_empty"], numeric_empty, strict=True
        )
        np.testing.assert_array_equal(mdict["char_empty"], char_empty, strict=True)
        np.testing.assert_array_equal(
            mdict["logical_empty"], logical_empty, strict=True
        )


@pytest.mark.parametrize("filename, version", files)
class TestSaveBasicDatatypes:

    def test_numeric_int8(self, filename, version):
        """Test writing int8 data to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["int8_scalar", "int8_array"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(
                temp_file_path, variable_names=["int8_scalar", "int8_array"]
            )

            np.testing.assert_array_equal(
                mdict["int8_scalar"], mload["int8_scalar"], strict=True
            )
            np.testing.assert_array_equal(
                mdict["int8_array"], mload["int8_array"], strict=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_numeric_uint8(self, filename, version):
        """Test writing uint8 data to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["uint8_scalar", "uint8_array"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(
                temp_file_path, variable_names=["uint8_scalar", "uint8_array"]
            )

            np.testing.assert_array_equal(
                mdict["uint8_scalar"], mload["uint8_scalar"], strict=True
            )
            np.testing.assert_array_equal(
                mdict["uint8_array"], mload["uint8_array"], strict=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_numeric_int16(self, filename, version):
        """Test writing int16 data to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["int16_scalar", "int16_array"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(
                temp_file_path, variable_names=["int16_scalar", "int16_array"]
            )

            np.testing.assert_array_equal(
                mdict["int16_scalar"], mload["int16_scalar"], strict=True
            )
            np.testing.assert_array_equal(
                mdict["int16_array"], mload["int16_array"], strict=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_numeric_uint16(self, filename, version):
        """Test writing uint16 data to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(
            file_path, variable_names=["uint16_scalar", "uint16_array"]
        )

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(
                temp_file_path, variable_names=["uint16_scalar", "uint16_array"]
            )

            np.testing.assert_array_equal(
                mdict["uint16_scalar"], mload["uint16_scalar"], strict=True
            )
            np.testing.assert_array_equal(
                mdict["uint16_array"], mload["uint16_array"], strict=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_numeric_int32(self, filename, version):
        """Test writing int32 data to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["int32_scalar", "int32_array"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(
                temp_file_path, variable_names=["int32_scalar", "int32_array"]
            )

            np.testing.assert_array_equal(
                mdict["int32_scalar"], mload["int32_scalar"], strict=True
            )
            np.testing.assert_array_equal(
                mdict["int32_array"], mload["int32_array"], strict=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_numeric_uint32(self, filename, version):
        """Test writing uint32 data to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(
            file_path, variable_names=["uint32_scalar", "uint32_array"]
        )

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(
                temp_file_path, variable_names=["uint32_scalar", "uint32_array"]
            )

            np.testing.assert_array_equal(
                mdict["uint32_scalar"], mload["uint32_scalar"], strict=True
            )
            np.testing.assert_array_equal(
                mdict["uint32_array"], mload["uint32_array"], strict=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_numeric_int64(self, filename, version):
        """Test writing int64 data to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["int64_scalar", "int64_array"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(
                temp_file_path, variable_names=["int64_scalar", "int64_array"]
            )

            np.testing.assert_array_equal(
                mdict["int64_scalar"], mload["int64_scalar"], strict=True
            )
            np.testing.assert_array_equal(
                mdict["int64_array"], mload["int64_array"], strict=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_numeric_uint64(self, filename, version):
        """Test writing uint64 data to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(
            file_path, variable_names=["uint64_scalar", "uint64_array"]
        )

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(
                temp_file_path, variable_names=["uint64_scalar", "uint64_array"]
            )

            np.testing.assert_array_equal(
                mdict["uint64_scalar"], mload["uint64_scalar"], strict=True
            )
            np.testing.assert_array_equal(
                mdict["uint64_array"], mload["uint64_array"], strict=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_numeric_single(self, filename, version):
        """Test writing single data to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(
            file_path, variable_names=["single_scalar", "single_array"]
        )

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(
                temp_file_path, variable_names=["single_scalar", "single_array"]
            )

            np.testing.assert_array_equal(
                mdict["single_scalar"], mload["single_scalar"], strict=True
            )
            np.testing.assert_array_equal(
                mdict["single_array"], mload["single_array"], strict=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_numeric_double(self, filename, version):
        """Test writing double data to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(
            file_path, variable_names=["double_scalar", "double_array"]
        )

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(
                temp_file_path, variable_names=["double_scalar", "double_array"]
            )

            np.testing.assert_array_equal(
                mdict["double_scalar"], mload["double_scalar"], strict=True
            )
            np.testing.assert_array_equal(
                mdict["double_array"], mload["double_array"], strict=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_logical(self, filename, version):
        """Test writing logical data to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(
            file_path, variable_names=["logical_scalar", "logical_array"]
        )

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(
                temp_file_path, variable_names=["logical_scalar", "logical_array"]
            )

            np.testing.assert_array_equal(
                mdict["logical_scalar"], mload["logical_scalar"], strict=True
            )
            np.testing.assert_array_equal(
                mdict["logical_array"], mload["logical_array"], strict=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_complex(self, filename, version):
        """Test writing complex data to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(
            file_path, variable_names=["complex_scalar", "complex_array"]
        )

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(
                temp_file_path, variable_names=["complex_scalar", "complex_array"]
            )

            np.testing.assert_array_equal(
                mdict["complex_scalar"], mload["complex_scalar"], strict=True
            )
            np.testing.assert_array_equal(
                mdict["complex_array"], mload["complex_array"], strict=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_char(self, filename, version):
        """Test writing char data to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["char_scalar", "char_array"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(
                temp_file_path, variable_names=["char_scalar", "char_array"]
            )

            np.testing.assert_array_equal(
                mdict["char_scalar"], mload["char_scalar"], strict=True
            )
            np.testing.assert_array_equal(
                mdict["char_array"], mload["char_array"], strict=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_empty(self, filename, version):
        """Test writing empty char data to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(
            file_path, variable_names=["char_empty", "numeric_empty", "logical_empty"]
        )

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(
                temp_file_path,
                variable_names=["char_empty", "numeric_empty", "logical_empty"],
            )

            np.testing.assert_array_equal(
                mdict["char_empty"], mload["char_empty"], strict=True
            )
            np.testing.assert_array_equal(
                mdict["numeric_empty"], mload["numeric_empty"], strict=True
            )
            np.testing.assert_array_equal(
                mdict["logical_empty"], mload["logical_empty"], strict=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)


@pytest.mark.parametrize("filename, version", files)
class TestWriteNonSupportedNumeric:

    def test_numpy_floats(self, filename, version):
        """Test writing numpy float16 data to MAT-file"""
        arr_16 = np.array([1, 2, 3], dtype=np.float16).reshape(1, 3)
        arr_128 = np.array([1, 2, 3], dtype=np.float128).reshape(1, 3)
        mdict = {"float16_array": arr_16, "float128_array": arr_128}

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            with pytest.warns(MatWriteWarning, match="not supported"):
                save_to_mat(temp_file_path, mdict, version=version)

            mload = load_from_mat(temp_file_path, variable_names=None)

            np.testing.assert_array_equal(
                mload["float16_array"],
                mdict["float16_array"].astype("float64"),
                strict=True,
            )
            np.testing.assert_array_equal(
                mload["float128_array"],
                mdict["float128_array"].astype("float64"),
                strict=True,
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_numpy_complex(self, filename, version):
        """Test writing numpy complex64 and complex256 data to MAT-file"""
        arr_c64 = np.array([1 + 2j, 3 + 4j, 5 + 6j], dtype=np.complex64).reshape(1, 3)
        arr_c256 = np.array([1 + 2j, 3 + 4j, 5 + 6j], dtype=np.complex256).reshape(1, 3)
        mdict = {"complex64_array": arr_c64, "complex256_array": arr_c256}

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            with pytest.warns(MatWriteWarning, match="not supported"):
                save_to_mat(temp_file_path, mdict, version=version)

            mload = load_from_mat(temp_file_path, variable_names=None)

            np.testing.assert_array_equal(
                mload["complex64_array"],
                mdict["complex64_array"].astype("complex128"),
                strict=True,
            )
            np.testing.assert_array_equal(
                mload["complex256_array"],
                mdict["complex256_array"].astype("complex128"),
                strict=True,
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)
